﻿Imports System.Runtime.CompilerServices
Imports System.ComponentModel, System.Diagnostics

Namespace System.Data.Common
	<EditorBrowsable(EditorBrowsableState.Never)>
	Public Module DataX
		<Extension()>
		Public Function Persistance(dts As DataSet) As DbPersistanceBase
			Return DbPersistanceBase.Get(dts)
		End Function
		<Extension()>
		Public Function TablePersist(table As DataTable) As TablePersist
			Return TablePersist.Get(table)
		End Function

		''' <summary> since SELECT and FROM-Clause are predefined by the datatable-structure, sqlAfterFrom can contain any other valid Sql-clauses, especially WHERE. To prevent Sql-Injection-attacs it is strongly recommended to use '?' as Parameter-PlaceHolder and submit appropriate args.</summary>
		<Extension()>
		Public Sub CustomFill(table As DataTable, sqlAfterFrom As String, ParamArray args() As Object)
			table.TablePersist.CustomFill("", sqlAfterFrom, args)
		End Sub

		''' <summary> since SELECT and FROM-Clause are predefined by the datatable-structure, sqlAfterFrom can contain any other valid Sql-clauses, especially WHERE. To prevent Sql-Injection-attacs it is strongly recommended to use '?' as Parameter-PlaceHolder and submit appropriate args. preamble may be "Distinct", "Top 100" and stuff</summary>
		<Extension()>
		Public Sub CustomFill(table As DataTable, preamble As String, sqlAfterFrom As String, args() As Object)
			table.TablePersist.CustomFill(preamble, sqlAfterFrom, args)
		End Sub

		''' <summary> fills childTable with those Rows, who reference rwParent by DataRelation. </summary>
		<Extension()>
		Public Sub FillChildTable(rwParent As DataRow, childTable As DataTable)
			Try
				Dim rl = rwParent.Table.ChildRelations.Cast(Of DataRelation).Single(Function(r) r.ChildTable Is childTable)
				rwParent.FillChildTable(rl)
			Catch ex As InvalidOperationException
				Throw New InvalidOperationException("a single DataRelation between Parent- and Child-Table is required", ex)
			End Try
		End Sub

		''' <summary> fills childTable with those Rows, who reference rwParent by DataRelation.</summary>
		<Extension()>
		Public Sub FillChildTable(rwParent As DataRow, rl As DataRelation)
			If rl.ParentTable IsNot rwParent.Table Then Throw New InvalidOperationException("the Relation doesn't reference the ParentTable")
			If rwParent.RowState = DataRowState.Detached Then
				rl.ChildTable.CustomFill(Nothing)
				Return
			End If
			Dim q = rl.DataSet.Persistance.Quotes
			Dim sqlChildCondition = $" where {String.Join(" AND ", rl.ChildColumns.Select(Function(x) $"{q(0)}{x.ColumnName}{q(1)} = ?"))}"
			rl.ChildTable.CustomFill(sqlChildCondition, rl.ParentColumns.Select(Function(x) rwParent(x)).ToArray)
		End Sub

		<Extension()>
		Public Sub FillAll(_Dts As DataSet)
			_Dts.Clear()
			_Dts.Persistance.RankedTables.ForEach(Sub(tb) tb.TablePersist.Fill())
		End Sub
		<Extension()>
		Public Sub Save(_Dts As DataSet)
			_Dts.Persistance.Save()
		End Sub
	End Module
End Namespace