﻿Imports System.Data.Common
Imports System.Diagnostics
Imports System.Text.RegularExpressions

<DebuggerDisplay("{_Table.TableName}-TablePersist")>
Public Class TablePersist : Inherits Runtime.CompilerServices.WeakAttachBase(Of TablePersist)

	Private _cmdCustomFill, _cmdInitialFill As DbCommand
	Public ReadOnly Adapter As DbDataAdapter
	Private Shared ReadOnly _rgxJoin As New Regex("^\W*(\w+\W*){2}join", RegexOptions.IgnoreCase)
	Private _tempTable As DataTable
	Public Event AfterFill As EventHandler

	Private _Table As DataTable
	Public ReadOnly Property Table() As DataTable
		Get
			Return _Table
		End Get
	End Property
	Public Sub New(table As DataTable, adapter As DbDataAdapter)
		_Table = table
		TablePersist.AddAttach(table, Me)
		_tempTable = table.DataSet.Persistance.DatasetClone.Tables(table.TableName)
		Me.Adapter = adapter
		_cmdInitialFill = adapter.SelectCommand
		_cmdCustomFill = DirectCast(DirectCast(_cmdInitialFill, ICloneable).Clone, DbCommand)
	End Sub

	Public Function Update(rows As DataRow()) As Integer
		Return Adapter.Update(rows)
	End Function
	Public Function Update() As Integer
		Return Adapter.Update(_Table)
	End Function
	Public Sub Fill()
		Dbg("Fill")
		FillAndRepair(_cmdInitialFill)
	End Sub
	Public Sub FillRefresh()
		Dbg("FillRefresh")
		FillAndRepair(_cmdCustomFill)
	End Sub
	''' <summary> since SELECT and FROM-Clause are predefined by the datatable-structure, sqlAfterFrom can contain any other valid Sql-clauses, especially WHERE. To prevent Sql-Injection-attacs it is strongly recommended to use '?' as Parameter-PlaceHolder and submit appropriate args. preamble may be "Distinct", "Top 100" and stuff</summary>
	Public Sub CustomFill(preamble As String, sqlAfterFrom As String, args As Object())
		_cmdCustomFill.Dispose()
		If sqlAfterFrom = "" Then ' soll nur gelöscht werden - kein DbZugriff
			_cmdCustomFill = Nothing
			FillAndRepair(_cmdCustomFill)
			Return
		End If
		_cmdCustomFill = DirectCast(DirectCast(_cmdInitialFill, ICloneable).Clone, DbCommand)
		Dim afterFromSplits = sqlAfterFrom.Split({"?"c}, StringSplitOptions.RemoveEmptyEntries) ' '?' durch benannte Parameter ersetzen
		For i = 0 To args.Length - 1
			Dim name = "@p" & i
			Dim prm = _cmdCustomFill.CreateParameter
			prm.Value = args(i)
			prm.ParameterName = name
			_cmdCustomFill.Parameters.Add(prm)
			afterFromSplits(i) &= name
		Next
		Dim selectSplits = _cmdCustomFill.CommandText.Split({" "c}, StringSplitOptions.RemoveEmptyEntries)
		selectSplits(0) &= " " & preamble ' preamble dem ersten Wort anfügen
		If _rgxJoin.IsMatch(sqlAfterFrom) Then
			selectSplits(selectSplits.Length - 1) = "" ' bei Join das letzte Wort entfernen, denn der Joining-Ausdruck benennt bereits die ZielTabelle
		End If
		_cmdCustomFill.CommandText = String.Join(" ", selectSplits.Concat(afterFromSplits))
		FillAndRepair(_cmdCustomFill)
	End Sub
	Private Sub ExecuteFill(cmd As DbCommand)
		'befüllt _Table aus der DB. Falls _Table nicht leer eine tempTable befüllen und in _Table eingemergen.
		Dim msg = {cmd.CommandText}.ToList
		Adapter.SelectCommand = cmd
		Dim loadeds = 0
		If _Table.Rows.Count = 0 Then
			loadeds = Adapter.Fill(_Table)
		Else
			loadeds = Adapter.Fill(_tempTable)
			_Table.Merge(_tempTable, False, missingSchemaAction:=MissingSchemaAction.Error)
			_tempTable.Clear()
		End If
		Dim line = $"{loadeds} Results"
		If cmd.Parameters.Count > 0 Then line &= "; Parameter(s): " & ", ".Between(cmd.ParametersX.Select(Function(x) x.Value))
		msg.Add(line)
		Debug.WriteLine(Lf.Between(msg))
		RaiseEvent AfterFill(Me, EventArgs.Empty)
	End Sub
	Private Sub FillAndRepair(cmd As DbCommand)
		'befülle _Table und repariere evtl. entstehende FK-Errors.
		'Zunächstmal in RepairUpwards() durch Refresh der ParentTables, die den jeweils FK-Error bedingen - cascadierend.
		'Dann durch Löschung der FK-Error-Rows in allen ChildTables der durch RepairUpwards refreshten Tables - auch cascadierend.
		'_Table.ClearCascade
		If cmd Is Nothing Then Return
		_cmdInitialFill.Connection.OpenUntilIdle
		_Table.DataSet.SuspendConstraintsUntilIdle
		Dim prevCount = _Table.Rows.Count
		ExecuteFill(cmd)
		Dim allvanished = prevCount > 0 AndAlso _Table.Rows.Count = 0
		Dim refresheds = If(allvanished, RepairUpwardsWhenAllVanished(_Table), RepairUpwards(_Table))
		RepairDownwards(refresheds)
	End Sub
	Private Shared Function RepairUpwardsWhenAllVanished(tbRoot As DataTable) As HashSet(Of DataTable)
		'Ein FloodFill - Algo entlang der Successor-Verknüpfung tb->tb.ParentTables - sofern die DataRelation in tb FK-Errors bedingt.
		'Als Repair werden die ParentTables refresht - und dann ist bei ihnen auch auf FK-Errors zu testen.
		'Die Successor-Verfolgung kann auch auf bereits behandelte Tables führen - diese natürlich nicht erneut behandeln - FloodFill eben.
		Dim refilledTables = New HashSet(Of DataTable) From {tbRoot}
		Dim qu = New Queue(Of DataTable)(refilledTables)
		While qu.Count > 0
			Dim tb = qu.Dequeue
			For Each tbParent In tb.ParentTables
				Dim persist = tbParent.TablePersist
				tbParent.Clear()
				persist.ExecuteFill(persist._cmdCustomFill)
				If tbParent.Rows.Count = 0 Then qu.Enqueue(tbParent)
			Next
		End While
		Return refilledTables
	End Function
	Private Shared Function RepairUpwards(tbRoot As DataTable) As HashSet(Of DataTable)
		'Ein FloodFill - Algo entlang der Successor-Verknüpfung tb->tb.ParentTables - sofern die DataRelation in tb FK-Errors bedingt.
		'Als Repair werden die ParentTables refresht - und dann ist bei ihnen auch auf FK-Errors zu testen.
		'Die Successor-Verfolgung kann auch auf bereits behandelte Tables führen - diese natürlich nicht erneut behandeln - FloodFill eben.
		Dim refilledTables = New HashSet(Of DataTable) From {tbRoot}
		Dim qu = New Queue(Of DataTable)(refilledTables)
		Dim parentTableRepairAction = Sub(tbParent As DataTable)
													If Not refilledTables.Add(tbParent) Then Return
													qu.Enqueue(tbParent)
													Dim persist = tbParent.TablePersist
													persist.ExecuteFill(persist._cmdCustomFill)
												End Sub
		While qu.Count > 0
			Dim tb = qu.Dequeue
			If HandleFK_Errors(tb, False, parentTableRepairAction) Then HandleFK_Errors(tb, False) ' bei FK-Error RepairAction-Ergebnis verifizieren
		End While
		Return refilledTables
	End Function
	Private Shared Sub RepairDownwards(refresheds As HashSet(Of DataTable))
		'FloodFill-Algo entlang der Successor-Verknüpfung tb->tb.ChildTables - sofern tb FK-Errors enthält.
		'Als Repair werden die Fehler-Rows einfach aus tb entfernt - und dann sind alle tb.ChildTables ebenfalls auf FK-Errors zu testen.
		'Die Successor-Verfolgung kann auch auf bereits behandelte Tables führen - diese natürlich nicht erneut behandeln - FloodFill eben.
		Dim qu = New Queue(Of DataTable)(refresheds)
		While qu.Count > 0
			For Each tb In qu.Dequeue.ChildTables.Where(AddressOf refresheds.Add)
				If HandleFK_Errors(tb, removeErrorRows:=True) Then qu.Enqueue(tb) ' bei FK-Error Suche vertiefen
			Next
		End While
	End Sub
	Private Shared Function HandleFK_Errors(tb As DataTable, removeErrorRows As Boolean, Optional parentTableRepairAction As Action(Of DataTable) = Nothing) As Boolean
		'Suche und behandle FK-Errors in tb - drei Betriebs-modi sind möglich:
		'1) If removeErrorRows Then tb.Rows.RemoveAt(iError)
		'2) Else If parentTableRepairAction Is Nothing Then Throw KeyNotFoundException
		'3) Else parentTableRepairAction.Invoke(tbParent)
		HandleFK_Errors = False
		Dim searchRelations = tb.ParentRelationsX
		For i = tb.Rows.Count - 1 To 0 Step -1 ' rückwärts durchlaufen, da währenddessen Löschungen stattfinden
			Dim rw = tb(i)
CheckRow:
			For Each rl In searchRelations
				Dim rwParent = rw.GetParentRow(rl)
				Dim clm = rl.ChildColumns(0)
				If rwParent.NotNull OrElse rw.IsNull(clm) Then Continue For ' rw(clm) ist OK
				' ab hier FK-Error: rwParent.Null, aber rw(clm).NotNull
				HandleFK_Errors = True
				If removeErrorRows Then tb.Rows.RemoveAt(i) : Continue For ' Reparatur durch row-Entfernung
				Dim tbParent = rl.ParentTable
				If parentTableRepairAction Is Nothing Then
					Dim errValue = rw(clm),
						childColFullname = $"{tb.TableName}.{clm.ColumnName}",
						parentColFullname = $"{tbParent.TableName}.{rl.ParentColumns(0).ColumnName}"
					Throw New KeyNotFoundException($"Value {childColFullname}={errValue} not found in {parentColFullname}")
				End If
				parentTableRepairAction.Invoke(tbParent)
				searchRelations.RemoveAll(Function(x) x.ParentTable Is tbParent)
				If searchRelations.Count = 0 Then Exit Function
				GoTo CheckRow
			Next
		Next
	End Function
End Class
