﻿Imports System.Runtime.CompilerServices
Imports System.Collections.Generic

Namespace Global.System.Data

	''' <summary>
	''' über Datarelations im Dataset navigieren - derzeit 3 Features:
	''' 1) GetRelationPathTo findet die DataRelations, die von einer Tabelle zur anderen führen
	''' 2) GetChildRelations findet alle DataRelations, die einer Tabelle untergeordnet sind
	''' 3) IsParentOf checkt, ob eine Tabelle einer anderen übergeordnet ist
	''' </summary>
	Public Module PathFinder

		Private _Goal As DataTable
		Private _EnforceUnique As Boolean
		'da _RelationPath vom Ziel her aufgebaut wird, spart ein Stack das Umkehren der Reihenfolge
		Private _RelationPath As Stack(Of DataRelation)
		Private _Collecting As List(Of DataRelation)
		Private _KnownSteps As New Stack(Of DataTable) 'um zu vermeiden, im Kreis zu laufen
		Private _GoalFound As Boolean

		''' <summary> enforceUnique: wirft Exception, wenn mehr als ein Pfad gefunden </summary>
		<Extension()>
		Public Function GetRelationPathTo(root As DataTable, goal As DataTable,
				Optional enforceUnique As Boolean = False) As Stack(Of DataRelation)
			EnterEnumeration(root.ChildRelations, goal, enforceUnique, New Stack(Of DataRelation), Nothing)
			Return _RelationPath
		End Function

		<Extension()>
		Public Function GetAllChildRelations(root As DataTable,
				Optional enforceUnique As Boolean = False) As List(Of DataRelation)
			EnterEnumeration(root, Nothing, enforceUnique, Nothing, New List(Of DataRelation))
			Return _Collecting
		End Function

		<Extension()>
		Public Function IsParentOf(root As DataTable, goal As DataTable) As Boolean
			EnterEnumeration(root, goal, True, Nothing, Nothing)
			Return _GoalFound
		End Function

		Private Sub EnterEnumeration(rels As DataRelationCollection, goal As DataTable,
				enforceUnique As Boolean, relationPath As Stack(Of DataRelation),
				collecting As List(Of DataRelation))
			_Goal = goal
			_EnforceUnique = enforceUnique
			_RelationPath = relationPath
			_Collecting = collecting
			_KnownSteps.Clear()
			_GoalFound = False
			Enumerate(rels)
		End Sub

		Private Sub Enumerate(rels As DataRelationCollection)
			'dieser rekursive Durchgang bedient 3 Aspekte:
			'entweder genau einen _RelationPath finden - wenns weitere gibt, Exception werfen
			'oder nur checken, ob irgendeine Verbindung existiert
			'oder alle _ChildRelations einsammeln
			For Each rl As DataRelation In rels
				If _Collecting IsNot Nothing Then _Collecting.Add(rl)
				Dim tbl = rl.ChildTable
				If tbl Is _Goal Then
					If _RelationPath IsNot Nothing Then
						If _RelationPath.Count > 0 Then Throw New InvalidOperationException("mehr als einen Pfad gefunden")
						_RelationPath.Push(rl)
					End If
					_GoalFound = True
					Return
				End If
				If _KnownSteps.Contains(tbl) Then Continue For
				_KnownSteps.Push(tbl)
				Enumerate(tbl.ChildRelations)
				If _GoalFound Then
					If _RelationPath IsNot Nothing Then _RelationPath.Push(rl)
					If Not _EnforceUnique Then Return
				End If
				_KnownSteps.Pop()
			Next
		End Sub

		Private Sub EnterEnumeration(root As DataTable, goal As DataTable,
				enforceUnique As Boolean, relationPath As Stack(Of DataRelation),
				collecting As List(Of DataRelation))
			_Goal = goal
			_EnforceUnique = enforceUnique
			_RelationPath = relationPath
			_Collecting = collecting
			_KnownSteps.Clear()
			_GoalFound = False
			Enumerate(root)
		End Sub

		Private Sub Enumerate(Table As DataTable)
			'dieser rekursive Durchgang bedient 3 Aspekte:
			'entweder genau einen _RelationPath finden - wenns weitere gibt, Exception werfen
			'oder nur checken, ob irgendeine Verbindung existiert
			'oder alle _ChildRelations einsammeln
			If _KnownSteps.Contains(Table) Then Return 'nicht im Kreis laufen
			If Table Is _Goal Then
				If _RelationPath IsNot Nothing AndAlso _RelationPath.Count > 0 Then
					Throw New InvalidOperationException("mehr als einen Pfad gefunden")
				End If
				_GoalFound = True
				Return
			End If
			_KnownSteps.Push(Table)
			For Each rl As DataRelation In Table.ChildRelations
				If _Collecting IsNot Nothing Then _Collecting.Add(rl)
				Enumerate(rl.ChildTable)
				If _GoalFound Then
					If _RelationPath IsNot Nothing Then _RelationPath.Push(rl)
					If Not _EnforceUnique Then Return
				End If
			Next
			_KnownSteps.Pop()
		End Sub

	End Module


End Namespace