﻿#Region "FileHeader"
#If False Then
WeakAttachBase(Of TAttach) ist eine Art statisches Dictionary(Of Object, TAttach), mit dem TAttach-Instanzen an beliebige Objekte "drangeklebt" werden können.
Die Funktionalität der Objekte ist so erweiterbar, ohne sie zu beerben.
Fehlende Einträge können auch auto-generiert werden beim Abruf eines Attaches - 2 Möglichkeiten: 
1) Aufruf eines parameterloser ctors erstellt das fehlende Attachment 
2) Eine Func(Of Object, TAttach) wird mitgegeben, welche aus dem Key das Attachment ableiten kann.
	Hierfür am besten im static ctor des Attachmets den geerbten statischen _Creater-Delegat initialisieren.
Ansonsten muss der Eintrag eben explizit eingetragen werden.
Kann das Attachment nicht von WeakAttachBase(Of TAttach) erben, so kann eine leere private AttachHelper-Class geschaffen werden, deren Member dann eben nach aussen geleitet werden. Siehe dazu letztes Beispiel im Tryals-Namespace
#End If '-- Options, Imports
#End Region 'FileHeader

#If True Then
Namespace Global.System.Runtime.CompilerServices

	Public MustInherit Class WeakAttachBase(Of TAttach As Class)
		Public Shared ReadOnly Attachs As New ConditionalWeakTable(Of Object, TAttach)

		Public Shared Function AddAttach(key As Object, value As TAttach) As TAttach
			Attachs.Add(key, value) : Return value
		End Function
		Public Shared Function RemoveAttach(key As Object) As Boolean
			Return Attachs.Remove(key)
		End Function
		''' <summary>gibt das dem key zugeordnete Attachment zurück - ist noch keines zugeordnet, wirds erstellt </summary>
		Public Shared Function [Get](key As Object) As TAttach
			Try
				Return Attachs.GetOrCreateValue(key)
			Catch ex As MissingMemberException
				Throw New KeyNotFoundException($"Key {key} not found - and couldn't create a new entry of type {GetType(TAttach)}", ex)
			End Try
		End Function
		Public Shared Function Exists(key As Object) As Boolean
			Dim value As TAttach = Nothing
			Return Attachs.TryGetValue(key, value)
		End Function

	End Class ' WeakAttachBase

	Public MustInherit Class WeakAttachBase(Of TKey As Class, TAttach As Class)

		Public Shared _Creater As ConditionalWeakTable(Of TKey, TAttach).CreateValueCallback
		Public Shared ReadOnly Attachs As New ConditionalWeakTable(Of TKey, TAttach)

		Shared Sub New()
			RuntimeHelpers.RunClassConstructor(GetType(TAttach).TypeHandle) ' des Erben TypeInitializer aufrufen - falls dort _Creater gesetzt wird
		End Sub

		Public Shared Function AddAttach(key As TKey, value As TAttach) As TAttach
			Attachs.Add(key, value) : Return value
		End Function
		Public Shared Function RemoveAttach(key As TKey) As Boolean
			Return Attachs.Remove(key)
		End Function

		''' <summary>gibt das dem key zugeordnete Attachment zurück - ist noch keines zugeordnet, wirds erstellt </summary>
		Public Shared Function [Get](key As TKey) As TAttach
			Try
				If _Creater Is Nothing Then Return Attachs.GetOrCreateValue(key)
				Return Attachs.GetValue(key, _Creater)
			Catch ex As MissingMemberException
				Throw New KeyNotFoundException($"Key {key} not found - and couldn't create a new entry of type {GetType(TAttach)}", ex)
			End Try
		End Function

		''' <summary>Vermeldet, wenn value neu erstellt wurde</summary>
		Public Shared Function CreateIfMissing(key As TKey, ByRef value As TAttach) As Boolean
			CreateIfMissing = Not Attachs.TryGetValue(key, value)
			If CreateIfMissing Then value = [Get](key)
		End Function
		Public Shared Function Exists(key As TKey) As Boolean
			Dim value As TAttach = Nothing
			Return Attachs.TryGetValue(key, value)
		End Function
	End Class ' WeakAttachBase

	Namespace Tryals
		'In den Beispielen sind clsKey und Attachment immer über Konstruktoren miteinander verbunden. Das ist nicht notwendig: jedes Dingens kann an jedes attachen
		Public Class clsKey
			Public P1 As String = "P1", P2 As String = "P2", P3 As String = "P3"
		End Class
		Public Class Attachment : Inherits WeakAttachBase(Of Attachment)
			Public Sub New() ' Auto-Erzeugung in BaseClass, per 0-param-ctor
			End Sub
			Public _Property1 As String, _Property2 As String = "Walt"
			Public Sub New(Property1 As String, Property2 As String)
				_Property1 = Property1 : _Property2 = Property2
			End Sub
		End Class

		Public Class Attachment2 : Inherits WeakAttachBase(Of clsKey, Attachment2)
			Shared Sub New()
				_Creater = Function(k) New Attachment2(k.P1, k.P2) ' Auto-Erzeugung in BaseClass per Ableitung aus dem Key
			End Sub
			Public _Property1, _Property2 As String
			Public Sub New(Property1 As String, Property2 As String)
				_Property1 = Property1 : _Property2 = Property2
			End Sub
		End Class
		Public Class Attachment3 : Inherits WeakAttachBase(Of Attachment3)
			Private Key As clsKey
			Public _Property1, _Property2 As String
			Public Sub New(Key As clsKey, Property1 As String, Property2 As String)
				Me.Key = Key : _Property1 = Property1 : _Property2 = Property2
				AddAttach(Key, Me)                ' Erzeugung des Eintrags in BaseClass, per ausdrücklicher Selbst-Zufügung
			End Sub
		End Class
		Public Class Attachment4
			'hier wird nicht von WeakAttach geerbt, sondern eine innere Attach-Klasse gebastelt. Deren Member - sofern gewünscht - müssen ausgeleitet werden
			Private Class Attaches : Inherits WeakAttachBase(Of Attachment4)
			End Class
			Public Shared Function [Get](key As clsKey) As Attachment4
				Return Attaches.Get(key)
			End Function
			Private Key As clsKey
			Public _Property1, _Property2 As String
			Public Sub New(Key As clsKey, Property1 As String, Property2 As String)
				Me.Key = Key : _Property1 = Property1 : _Property2 = Property2
				Attaches.AddAttach(Key, Me)               ' Erzeugung des Eintrags in BaseClass, per ausdrücklicher Selbst-Zufügung
			End Sub
		End Class

		Public Class Tester
			Public Shared Sub Test()
				Dim k = New clsKey, k2 = New clsKey With {.P2 = "Wand"}
				Dim x = Attachment.Get(k)
				Dim xx = Attachment2.Get(k)
				Dim xy = Attachment2.Get(k2)
				Call New Attachment3(k, "Hallo", "Welt").Nop
				Dim xxx = Attachment3.Get(k)
				Call New Attachment4(k, "Hallo", "Welt").Nop
				Dim xxxx = Attachment4.Get(k)
				Nops.Nop()
			End Sub
		End Class
	End Namespace

End Namespace
#ElseIf True Then
Namespace Global.System.Runtime.CompilerServices

	''' <summary> Erben dieser Klasse sind mit beliebigen Daten-Feldern und whatever zu erweitern. Wichtigste Vererbung ist die Shared Get - Methode, mit der ein Attachment (selbstinstanzierend) von beliebigen Objekten abgerufen werden kann. </summary>
	Public MustInherit Class WeakAttachBase(Of Tself As {WeakAttachBase(Of Tself), New})

		Protected Shared _Creater As ConditionalWeakTable(Of Object, Tself).CreateValueCallback
		Public Shared ReadOnly Attachs As ConditionalWeakTable(Of Object, Tself) = New ConditionalWeakTable(Of Object, Tself)

		Shared Sub New()
			Dim dummy = New Tself 'Bug: Nur Instanzieren eines Dummies erzwingt zuverlässig den Aufruf von Shared Sub New() erbender Klassen. Ist nötig, damit dort ggfs ein _Creater gesetzt wern kann
			If _Creater Is Nothing Then _Creater = Function(key) New Tself
		End Sub

		''' <summary>gibt das dem key zugeordnete Attachment zurück - ist noch keines zugeordnet, wirds erstellt </summary>
		Public Shared Function [Get](key As Object) As Tself
			Return Attachs.GetValue(key, _Creater)
		End Function

		''' <summary>setzt value auf das dem key zugeordnete Attachment zurück - ist noch keines zugeordnet, wirds erstellt und returnt zusätzlich True </summary>
		Public Shared Function CreateIfMissing(key As Object, ByRef value As Tself) As Boolean
			CreateIfMissing = Not Attachs.TryGetValue(key, value)
			If CreateIfMissing Then
				value = _Creater(key)
				Attachs.Add(key, value)
			End If
		End Function
		Public Shared Function Exists(key As Object) As Boolean
			Dim value As Tself = Nothing
			Return Attachs.TryGetValue(key, value)
		End Function
	End Class

	''' <summary> Dieses Attach kann nur an TKey attached werden, und hält ausserdem einen Verweis auf den Key, an den es attached ist </summary>
	Public MustInherit Class WeakAttacCoupled(Of TKey As Class, Tself As {WeakAttacCoupled(Of TKey, Tself), New})

		Private Shared ReadOnly _IListFields As New ConditionalWeakTable(Of TKey, Tself)

		Protected Shared _Creater As ConditionalWeakTable(Of TKey, Tself).CreateValueCallback =
			Function(k) New Tself With {._Key = k}

		Protected _Key As TKey

		Public Shared Function GetAttached(key As TKey) As Tself
			Return _IListFields.GetValue(key, _Creater)
		End Function

	End Class

End Namespace
#ElseIf True Then
#End If

