﻿Imports System.Collections.Generic
Imports System.Linq.Expressions

Namespace Global.System.Reflection

	''' <summary> encapsulates 2 Delegates for read/write access a member of TOwner. Either set them manually or let Reflection derive them from Member-Name. In latter case also private Members are accessible, and the accesses are upto 30 times faster than original reflection </summary>
	Public Class MemberAccess(Of TOwner, TMember)
		Public ReadOnly Getter As Func(Of TOwner, TMember)
		Public ReadOnly Setter As Action(Of TOwner, TMember)
		Public ReadOnly Name As String
		Public Sub New(getter As Func(Of TOwner, TMember), Optional setter As Action(Of TOwner, TMember) = Nothing, Optional memberName As String = Nothing)
			Me.Getter = getter
			Me.Setter = setter
			Me.Name = memberName
		End Sub

		''' <summary> note: Instead of Exception on readonly-members the Setter-Initialization will be omitted </summary>
		Public Sub New(memberName As String, Optional friendlyName As String = Nothing)
			Me.Name = If(friendlyName, memberName)
			Dim tpOwner = GetType(TOwner)
			Dim flags = BindingFlags.Instance Or BindingFlags.[Public] Or BindingFlags.NonPublic
			Dim mmb As MemberInfo = tpOwner.GetProperty(memberName, flags)
			If mmb Is Nothing Then mmb = tpOwner.GetField(memberName, flags)
			If mmb Is Nothing Then _
				Throw New KeyNotFoundException(String.Format("member '{0}.{1}' not found", tpOwner.Name, memberName))
			Dim ownerParamX As ParameterExpression = Expression.Parameter(tpOwner, "owner")
			Dim memberParamX As ParameterExpression = Expression.Parameter(GetType(TMember), "value")
			Dim mmbX As MemberExpression = Expression.MakeMemberAccess(ownerParamX, mmb)
			Getter = Expression.Lambda(Of Func(Of TOwner, TMember))(mmbX, "get" & memberName, True, {ownerParamX}).Compile()
			Try
				Dim assignX = Expression.Assign(mmbX, memberParamX)
				Setter = Expression.Lambda(Of Action(Of TOwner, TMember))(assignX, "set" & memberName, True, {ownerParamX, memberParamX}).Compile()
			Catch
			End Try
		End Sub
		Default Public Property Item(owner As TOwner) As TMember
			Get
				Return Getter(owner)
			End Get
			Set(value As TMember)
				Setter(owner, value)
			End Set
		End Property
	End Class


End Namespace