﻿Imports System.ComponentModel

Namespace Global.System


	''' <summary>
	''' Mechanismus, um zu konfigurieren, daß die Property eines Items die Property eines anderen Items benachrichtigt. Die Synchronisation kann one-way (Notify) oder gegenseitig (Synchronisize) erfolgen. Optional können für die zu übermittelnden Werte Func(Of TIn, TOut) angegeben werden.
	''' </summary>
	Public MustInherit Class PropertyNotifyBase

		Protected MustOverride Function GetItem() As Object
		Protected MustOverride Function GetProp() As PropertyDescriptor
		Private Shared _OnInitPropNotify As PropertyNotifyBase

		Public Function OnInit() As PropertyNotifyBase
			_OnInitPropNotify = Me
			Return Me
		End Function

		Public Function Notify(other As Object, prop As String) As PropertyNotifyBase
			Return Notify(Of Object, Object)(other, prop, Nothing)
		End Function

		Public Function Notify(Of TIn, TOut)(
				 other As Object, prop As String,
				 converter As Func(Of TIn, TOut)) As PropertyNotifyBase
			Dim otherProp = TypeDescriptor.GetProperties(other)(prop)
			Return InitNotify(GetItem, GetProp, other, otherProp, converter, Nothing, True)
		End Function

		Public Function Synchronisize(other As Object, prop As String) As PropertyNotifyBase
			Return Synchronisize(Of Object, Object)(other, prop, Nothing)
		End Function

		Public Function Synchronisize(Of TIn, TOut)(other As Object, prop As String, converter As Func(Of TIn, TOut), Optional backConverter As Func(Of TOut, TIn) = Nothing) As PropertyNotifyBase
			Dim otherProp = TypeDescriptor.GetProperties(other)(prop)
			Return InitNotify(GetItem, GetProp, other, otherProp, converter, backConverter, False)
		End Function

		Public Function Notify(Of T)(other As NotifyProperty(Of T)) As PropertyNotifyBase
			Return Notify(Of T, Object, Object)(other, Nothing)
		End Function

		Public Function Notify(Of T, TIn, TOut)(
				 other As NotifyProperty(Of T),
				 converter As Func(Of TIn, TOut)) As PropertyNotifyBase
			Return InitNotify(GetItem, GetProp, other, other.GetProp, converter, Nothing, True)
		End Function

		Public Function Synchronisize(Of T2)(other As NotifyProperty(Of T2)) As PropertyNotifyBase
			Return Synchronisize(Of T2, Object, Object)(other, Nothing)
		End Function

		Public Function Synchronisize(Of T2, TIn, TOut)(other As NotifyProperty(Of T2), converter As Func(Of TIn, TOut), Optional backConverter As Func(Of TOut, TIn) = Nothing) As PropertyNotifyBase
			Return InitNotify(Me, GetProp, other, other.GetProp, converter, backConverter, False)
		End Function

		Private Function InitNotify(Of TIn, TOut)(
				 src As Object, prpSrc As PropertyDescriptor,
				 other As Object, otherProp As PropertyDescriptor,
				 customConverter As Func(Of TIn, TOut),
				 backConverter As Func(Of TOut, TIn),
				 oneWay As Boolean) As PropertyNotifyBase
			If _OnInitPropNotify IsNot Me Then _OnInitPropNotify = Nothing
			SetListener(src, prpSrc, other, otherProp, customConverter)
			If oneWay Then Return Me
			If backConverter Is Nothing Then Return SetListener(other, otherProp, src, prpSrc, customConverter)
			Return SetListener(other, otherProp, src, prpSrc, backConverter)
		End Function

		Protected Function SetListener(Of TIn, TOut)(
				 src As Object, prpSrc As PropertyDescriptor,
				 listener As Object, prpDest As PropertyDescriptor,
				 customConverter As Func(Of TIn, TOut)) As PropertyNotifyBase
			'Beim Verknüpfen von src und listener ist evtl. ein geeigneter Konverter zu suchen
			'(Die kunstvoll komponierten Fehlermeldungen machen diesen Code leider recht unleserlich)
			If Not prpSrc.SupportsChangeEvents Then
				Throw New ArgumentException(String.Concat(
					src.GetType.Name, " implementiert nicht INotifyPropertyChanged, und bietet auch kein der ",
					".", prpSrc.Name, "-Property entsprechendes ", prpSrc.Name, "Changed-Event."))
			End If
			Dim getNewValue As Func(Of Object, Object) = Nothing
			Dim tps = New Type() {prpSrc.PropertyType, prpDest.PropertyType}
			Dim msgHalfError = String.Concat(", um die Property ",
						src.GetType.Name, ".", prpSrc.Name, " (", tps(0).Name, ") in die Property ",
						listener.GetType.Name, ".", prpDest.Name, " (", tps(1).Name, ") zu konvertieren")
			If customConverter IsNot Nothing Then
				'ein benutzerdefinierte Konverter hat natürlich Vorrang
				If GetType(TIn).IsAssignableFrom(tps(0)) AndAlso tps(1).IsAssignableFrom(GetType(TOut)) Then
					getNewValue = Function(o) customConverter(DirectCast(o, TIn))
				Else
					Throw New InvalidCastException(String.Concat(
						Me.GetType.Name, ": Der benutzerdefinierte Func(Of ",
						GetType(TIn).Name, ", ", GetType(TOut).Name, ") ist nicht geeignet", msgHalfError))
				End If
			Else
				If tps(1).IsAssignableFrom(tps(0)) Then
					'Typen kompatibel - keine Konvertierung erf.
				ElseIf prpDest.Converter.CanConvertFrom(tps(0)) Then
					'Konverter der Ziel-Property anwenden
					getNewValue = AddressOf prpDest.Converter.ConvertFrom
				ElseIf prpSrc.Converter.CanConvertTo(tps(1)) Then
					'Konverter der Source-Property anwenden
					getNewValue = Function(o) prpSrc.Converter.ConvertTo(o, tps(1))
				Else
					Throw New InvalidCastException(String.Concat(
						Me.GetType.Name, " hat keinen geeigneten Konverter zur Verfügung", msgHalfError))
				End If
			End If
			'des Pudels Kern: Ein EventHandler wird gebildet, der den (evtl. konvertierten) Wert der Source-Property in die Destination-Property schreibt.
			Dim Handler As EventHandler =
				Function(sender, e) PushData(sender, prpSrc, getNewValue, listener, prpDest)
			'diesen EventHandler im Source-Objekt registrieren
			prpSrc.AddValueChanged(src, Handler)
			If _OnInitPropNotify Is Me Then Handler(src, EventArgs.Empty) 'ggfs erstmalige Synchronisation
			Return Me
		End Function

		Private Shared _senders As New Stack(Of Object())(New Object()() {New Object() {Nothing, Nothing}})

		Private Shared Function PushData(
				 src As Object, prpSrc As PropertyDescriptor,
				 getNewValue As Func(Of Object, Object),
				 dst As Object, prpDest As PropertyDescriptor) As Object
			Dim sender = _senders.Peek
			If sender(0) Is dst AndAlso sender(1) Is prpDest Then Return Nothing
			Dim newVal = prpSrc.GetValue(src)
			If getNewValue IsNot Nothing Then newVal = getNewValue(newVal)
			If Object.Equals(newVal, prpDest.GetValue(dst)) Then Return Nothing
			_senders.Push(New Object() {src, prpSrc})
			prpDest.SetValue(dst, newVal)
			_senders.Pop()
			Return Nothing
		End Function

	End Class 'PropertyNotifyBase
End Namespace