﻿Option Compare Text

Imports System.Runtime.CompilerServices
Imports System.Text
Imports System.Collections
Imports System.Text.RegularExpressions
Imports System.Globalization
Imports System.ComponentModel
Imports System.Diagnostics

Namespace Global.System

    Public Module StringX
        <Extension()>
        Public Sub AddLine(lines As List(Of String), ParamArray args As Object())
            lines.Add(String.Join(" ", args))
        End Sub

        Private _Counter, _MaxCount As Integer
        Private _patterns As Char() = "n,:0123456789".ToCharArray
        Private Function FormatXEvaluater(mt As Match) As String
            Dim s = mt.Value
            Dim i = s.IndexOfAny(_patterns, 1, 1)
            If i > 0 Then 'einer der _patterns matcht
                Select Case Array.IndexOf(_patterns, s(i))
                    Case 0 : Return Lf 'n matcht
                    Case Is < 3 'einer der beiden FormatString-Segmentierer
                    Case Else : Return s 'eine Ziffer
                End Select
            End If
            '_Counter als Ziffer einfügen
            If _Counter < _MaxCount Then _Counter += 1
            Return s.Insert(1, _Counter.ToString)
        End Function
        ''' <summary> funzt wie String.Format, nur wird in leere '{}' laufende Nummern eingetragen - jedoch nicht höhere als Anzahl args. Sodass das letzte arg auch vielfach eingefügt werden kann </summary>
        <Extension()>
        Public Function FormatX(sFormat As String, ParamArray args As Object()) As String
            Return Format(CultureInfo.CurrentCulture, sFormat, args)
        End Function
        ''' <summary> funzt wie String.Format, nur wird in leere '{}' laufende Nummern eingetragen - jedoch nicht höhere als Anzahl args. Sodass das letzte arg auch vielfach eingefügt werden kann </summary>
        <Extension()>
        Public Function FormatInvariant(sFormat As String, ParamArray args As Object()) As String
            Return Format(CultureInfo.InvariantCulture, sFormat, args)
        End Function
        Private Function Format(cult As CultureInfo, sFormat As String, ParamArray args As Object()) As String
            _Counter = -1 : _MaxCount = args.Length - 1
            'der regex identifiziert unescapte '{}' - ignoriert zB '{{}' oder '{}}', findet aber '{{{}'
            Dim s = Regex.Replace(sFormat, "(?<=([^{]|^)({{)*){[^{}]*[}]", AddressOf FormatXEvaluater)
            Return String.Format(cult, s, args)
        End Function

        ''' <summary> positive value cuts from left, negative cuts from right </summary>
        <Extension()>
        Public Function SubstringX(Subj As String, cut As Integer) As String
            If cut < 0 Then Return Subj.Substring(0, Subj.Length + cut)
            Return Subj.Substring(cut)
        End Function
        ''' <summary> start and end instead of start and length. Adds negative numbers to Subj.Length </summary>
        <Extension()>
        Public Function SubstringX(Subj As String, start As Integer, [end] As Integer) As String
            If start < 0 Then start += Subj.Length
            If [end] <= 0 Then [end] += Subj.Length
            Return Subj.Substring(start, [end] - start)
        End Function

        ''' <summary> more comfortable than String.Split(separator as String(),count,mode) </summary>
        <Extension()>
        Public Function Split(Subj As String, separator As String, count As Integer, Optional mode As StringSplitOptions = StringSplitOptions.None) As String()
            Return Subj.Split(New String() {separator}, count, mode)
        End Function
        <Extension()>
        Public Function Replace(Subj As String, oldValue As String, ParamArray concatNewValues() As Object) As String
            Return Subj.Replace(oldValue, String.Concat(concatNewValues))
        End Function

        ''' <summary> more comfortable to use than String.Split(separator as String(),mode) </summary>
        <Extension()>
        Public Function Split(Subj As String, separator As String, Optional mode As StringSplitOptions = StringSplitOptions.None) As String()
            Return Subj.Split(New String() {separator}, mode)
        End Function

        ''' <summary> verkettet alles </summary>
        <Extension()>
        Public Sub [Set](ByRef Subj As String, ParamArray others() As Object)
            Subj = String.Concat(others)
        End Sub

        <EditorBrowsable(EditorBrowsableState.Never)>
        Public Function ReplaceBslashN(s As String) As String
            Dim sb As New StringBuilder
            For i = 0 To s.Length - 2 Step 2
                If s.IndexOf("\n", i, 2) < 0 Then Exit For
                sb.Append(CrLf)
            Next
            sb.Append(s.Substring(sb.Length))
            Return sb.ToString
        End Function

        '''<summary> verkettet alles, interpretiert \n am Anfang als newline</summary>
        <Extension()>
        Public Function And2(Subj As String, ParamArray others() As Object) As String
            Dim sb = New StringBuilder(ReplaceBslashN(Subj))
            For Each itm In others
                sb.Append(ReplaceBslashN(itm.ToString))
            Next
            Return sb.ToString
        End Function

        '''<summary> verkettet alles</summary>
        <Extension()>
        Public Function [And](Subj As String, ParamArray others() As Object) As String
            Return String.Concat(Subj, String.Concat(others))
        End Function

        ''' <summary>Stringabschnitt ab sMatch's letzten Auftreten - optional einschließlich des Matches</summary>
        <Extension()>
        Public Function LastRightCut(s As String, sMatch As String, Optional includePattern As Boolean = False) As String
            Dim i = s.LastIndexOf(sMatch)
            If i < 0 Then Return Nothing
            Return s.Substring(If(includePattern, i, i + sMatch.Length))
        End Function

        ''' <summary>Stringabschnitt ab sMatch's erstem Auftreten - optional einschließlich des Matches</summary>
        <Extension()>
        Public Function RightCut(S$, sMatch$, Optional includePattern As Boolean = False) As String
            Dim I% = S.IndexOf(sMatch)
            If I < 0 Then Return Nothing
            Return S.Substring(If(includePattern, I, I + sMatch.Length))
        End Function

        Public Function CreateUnescapeRegex(Pattern As String) As Regex
            Return New Regex("(?<=(^|[^\\](\\\\)*))" & Pattern) 'der vorangestellte Pattern schließt escaped Matches ('\X' etc.) aus
        End Function

        <Extension(), DebuggerStepThrough()>
        Public Function HasValue(Subj As System.String) As Boolean
            Return Not String.IsNullOrEmpty(Subj)
        End Function

        <Extension(), DebuggerStepThrough()>
        Public Function HasNoValue(Subj As String) As Boolean
            Return String.IsNullOrEmpty(Subj)
        End Function

        ''' <summary>gibt den String-Abschnitt links des letzten gefundenen Matches zurück - optional einschließlich des Matches</summary>
        <Extension()>
        Public Function LastLeftCut(Subj As String, Pattern As String, Optional includePattern As Boolean = False) As String
            Dim i = Subj.LastIndexOf(Pattern)
            Return If(i >= 0, Subj.Substring(0, If(includePattern, i + Pattern.Length, i)), Nothing)
        End Function

        ''' <summary> gibt den String-Abschnitt links des ersten gefundenen Matches zurück - optional einschließlich des Matches</summary>
        <Extension()>
        Public Function LeftCut(Subj As String, Pattern As String, Optional includePattern As Boolean = False) As String
            Dim i = Subj.IndexOf(Pattern)
            Return If(i >= 0, Subj.Substring(0, If(includePattern, i + Pattern.Length, i)), Nothing)
        End Function

        Public Enum Additional : None : Start : [End] : Start_End : End Enum
        <Extension()>
        Public Function Between(separator As Char, additional As Additional, ParamArray args() As Object) As String
            Return Between(args, separator, additional)
        End Function
        <Extension()>
        Public Function Between(separator As Char, ParamArray args() As Object) As String
            Return Between(args, separator, Additional.None)
        End Function
        <Extension()>
        Public Function Between(separator As String, additional As Additional, ParamArray args() As Object) As String
            Return Between(args, separator, additional)
        End Function
        <Extension()>
        Public Function Between(separator As String, ParamArray args() As Object) As String
            If args.Length = 1 Then
                'special: args is just 2 chars in a string - treat the chars as opening and closing bracket or Quotes
                Dim s As String = TryCast(args(0), String)
                If s?.Length = 2 Then Return String.Concat(s(0), separator, s(1))
            End If
            Return Between(args, separator, Additional.None)
        End Function

        Private _SB As New System.Text.StringBuilder
#If True Then
        Private Function Between(args() As Object, separator As String, additional As Additional) As String
            While args.Length = 1 AndAlso TypeOf args(0) IsNot String
                Dim a = TryCast(args(0), IEnumerable)
                If a Is Nothing Then Exit While
                args = a.Cast(Of Object).ToArray
            End While
            _SB.Clear()
            If additional.HasFlag(Additional.Start) Then _SB.Append(separator)
            _SB.Append(String.Join(separator, args))
            If additional.HasFlag(Additional.End) Then _SB.Append(separator)
            Return _SB.ToString
        End Function
#ElseIf True Then
		'EA 2020-09: was before
		Private Function Between(args() As Object, separator As String, additional As Additional) As String
			Try
				While args.Length < 2 AndAlso Not TypeOf args(0) Is String
					args = DirectCast(args(0), IEnumerable).Cast(Of Object).ToArray
				End While
			Catch ex As InvalidCastException
			Catch ex As IndexOutOfRangeException ' can be InvalidCast or IndexOutOfRange
			End Try
			_SB.Clear()
			If additional.HasFlag(Additional.Start) Then _SB.Append(separator)
			_SB.Append(String.Join(separator, args))
			If additional.HasFlag(Additional.End) Then _SB.Append(separator)
			Return _SB.ToString
		End Function
#ElseIf True Then
#End If

        <Extension()>
        Public Function CountChars(S As String, Search As Char) As Integer
            CountChars = 0
            Dim I As Integer = S.IndexOf(Search)
            While I >= 0
                CountChars += 1
                I = S.IndexOf(Search, I + 1)
            End While
        End Function

        <Extension()>
        Public Function CountSubStrings(S As String, Search As String,
                Optional Comparison As StringComparison = StringComparison.CurrentCulture) As Integer
            CountSubStrings = 0
            Dim I As Integer = S.IndexOf(Search, 0, Comparison)
            While I >= 0
                CountSubStrings += 1
                I = S.IndexOf(Search, I + 1, Comparison)
            End While
        End Function

        <Extension()>
        Public Function TryParseX(s As String, defaultValue As Integer) As Integer
            Dim val As Integer : Return If(Integer.TryParse(s, val), val, defaultValue)
        End Function
        <Extension()>
        Public Function TryParseX(s As String, defaultValue As Double) As Double
            Dim val As Double : Return If(Double.TryParse(s, val), val, defaultValue)
        End Function
        <Extension()>
        Public Function TryParseX(s As String, defaultValue As Date) As Date
            Dim val As Date : Return If(Date.TryParse(s, val), val, defaultValue)
        End Function
        <Extension()>
        Public Function TryParseX(s As String, defaultValue As TimeSpan) As TimeSpan
            Dim val As TimeSpan : Return If(TimeSpan.TryParse(s, val), val, defaultValue)
        End Function

    End Module


End Namespace