﻿#Region "FileHeader"
#If False Then ' -------------- FileDescription --------------
temporäre Suspendierung von mit der DataTable zusammenhängender Funktionalität, in verschiedenen Stärken, entsprechend der Suspend-Enumeration:
NotfyOnly: nur das Suspending-Event raisen, etwa um Business-Logic zu deaktivieren
BindingSources: zusätzlich alle zu diesem Zwecke registrierten BindingSources deaktivieren
TableIndicees: zusätzlich mittels Table.BeginLoadData TableIndicees und Constraint-Überwachung deaktivieren
Die Suspendierung ist schachtelbar: wenn zB. übergeordneter Code alle Tabellen des Datasets suspendiert, und untergeordneter Code suspendiert nochmal einzelne Tabellen, so endet die Suspendierung erst mit Ende der übergeordneten
Bei Einschachtelung stärkerer Suspendierung wird "nach-suspendiert" 
Der Trick mit BeginSuspend() As IDisposable ermöglicht die Suspendierung im Using-Block.
Die Dataset.Register(Form)-Extension registriert sämtliche per Reflection aufgefundenen Bindingsources des Forms als suspendable
#End If
#End Region 'FileHeader

Namespace Global.System.Data

	Public Enum Suspendstate As Integer : Ended = -1 : BeforeEnd : Started : End Enum

	Partial Public Class DataTableAttach : Implements IDisposable
		Public Enum Suspend : None : NotifyOnly : BindingSources : TableIndicees : End Enum
		Private _Level As Integer
		Private _Suspend As Suspend
		Private _GroupSuspended As Boolean = False
		Friend ReadOnly SuspendableBindingSources As New List(Of BindingSource)
		Public Sub AddSource(bs As BindingSource)
			If SuspendableBindingSources.Contains(bs) Then Return
			SuspendableBindingSources.Add(bs)
			If _Level > 0 Then bs.RaiseListChangedEvents = False
			AddHandler bs.Disposed, AddressOf SuspendableBindingSource_Disposed
		End Sub

		Private Sub SuspendableBindingSource_Disposed(sender As Object, e As EventArgs)
			SuspendableBindingSources.Remove(DirectCast(sender, BindingSource))
		End Sub

		Public Function Enable() As Boolean
			Return _Suspend = Suspend.None
		End Function

		Public Sub SuspendLogic(value As Boolean, what As Suspend)
#If False Then
Schachtelbarkeit wird durch den Level-Zähler erreicht, der Suspendings hochzählt, und Resumings runter, und erst bei Level=0 wird die Resume-Action ausgeführt. 
Die Suspendierung ist aber **nicht** Level-abhängig, sondern, wenn eine stärkere Suspendierung angefordert wird als die bestehende, dann wird sie ausgeführt - ob nun auf Level=0 oder höher.
Einzig die Started-Notifikation wird ausschließlich bei Level=0 initial geraist. 
Effekt ist, dass eine Suspendierung mit Suspend.Notify starten kann, dann aber in geschachtelten Aufrufen verstärkt werden, sodass sie am Ende der äusseren Suspendierung ein Resume für eine Suspend-Stärke ausführt, mit der sie gar nicht begann.
#End If
			If value Then 'Suspending
				If what > _Suspend Then 'Bei stärkerer Suspendierung "nach-suspendiert" 
					_Suspend = what
					If _Suspend >= Suspend.TableIndicees Then
						Dataset_Attach.SuspendConstraints.Enter()
						Table.BeginLoadData()
					End If
					If _Suspend >= Suspend.BindingSources Then
						For Each bs In Me.SuspendableBindingSources : bs.RaiseListChangedEvents = False : Next
					End If
				End If
				If _Level = 0 Then RaiseSuspendingEvent(Suspendstate.Started, _Suspend)
				_Level += 1
			Else 'Resuming
				_Level -= 1
				If _Level = 0 Then ' execute Resume
					If Not _GroupSuspended Then RaiseSuspendingEvent(Suspendstate.BeforeEnd, _Suspend)
					If _Suspend >= Suspend.TableIndicees Then
						Table.EndLoadData()
						_Dataset_Attach.SuspendConstraints.Exit()
					End If
					If _Suspend >= Suspend.BindingSources Then
						For Each bs In Me.SuspendableBindingSources
							bs.RaiseListChangedEvents = True
							Try
								bs.ResetBindings(False)
							Catch ex As InvalidOperationException
								Application.OpenForms(0).BeginInvoke(Sub() bs.ResetBindings(False))
							End Try
						Next
					End If
					what = _Suspend
					_Suspend = Suspend.None
					RaiseSuspendingEvent(Suspendstate.Ended, what)
				End If
			End If
		End Sub

		Private Sub RaiseSuspendingEvent(state As Suspendstate, what As Suspend)
			Dim e = New SuspendingEventArgs(state, what)
			RaiseEvent Suspending(Me, e)
			If _Dataset_Attach.LogSuspendStates.Level > 0 Then Dbg("SuspendLogic", e.Suspendstate, Table.TableName, e.Suspended)
		End Sub

		Public Function BeginSuspend(what As Suspend) As IDisposable
			SuspendLogic(True, what)
			Return Me
		End Function
		Public Sub EndSuspend() Implements IDisposable.Dispose
			SuspendLogic(False, Suspend.None)
		End Sub

	End Class

End Namespace
