﻿#Region "FileHeader"
#If False Then ' -------------- FileDescription --------------
DataTable.Attach: returnt das zugehörige DataTableAttach
DataTable.Persist(Boolean): setzt Persistenz der DataTable - alle anneren wern gegenteilig gesetzt
DataSet.Persist(Boolean, DataTable()): setzt Persistenz der aufgeführten DataTables - alle anneren wern gegenteilig gesetzt
DataSet.PersistAll: setzt Persistenz aller DataTables auf True
TypedTableBase.Current: returnt/setzt eine Datarow als Current
DataTable.SetParentNotifications: setzt Current-Mitführung übergeordneter DataTableAttaches
BindingSource.RegisterCurrency: setzt Current-Synchronisierung mit einer BindingSource
DataSet.RegisterCurrency(BindingSource()): setzt Current-Synchronisierung  der aufgeführten BindingSources
DataSet.SuspendConstraints(): ermöglicht Constraints in einem Using-Block zu suspendieren
#End If
' -------------- Imports & Options --------------------
Imports System.Runtime.CompilerServices
Imports System.ComponentModel
#End Region 'FileHeader

Namespace Global.System.Data

	Public Module XDataTableAttach

		Private _TableAttachs As New ComponentProperty(Of DataTable, DataTableAttach)(Function(tb) New DataTableAttach(tb))
		''' <summary> holds additional functionality for the DataTable, like current-tracking, event-suspending </summary>
		<Extension(), DebuggerStepThrough()>
		Public Function Attach(tb As DataTable) As DataTableAttach
			Return _TableAttachs(tb)
		End Function
		''' <summary>mark this Table as to persist. All other tables will behave contrary. This call isn't necessary, but optimizes startup</summary>
		<Extension()>
		Public Sub Persist(tb As DataTable, value As Boolean)
			tb.DataSet.Persist(value, tb)
		End Sub
		''' <summary>mark these Tables as to persist. All other tables will behave contrary. This call isn't necessary, but optimizes startup</summary>
		<Extension()>
		Public Sub Persist(dts As DataSet, value As Boolean, ParamArray tables As DataTable())
			If DatasetAdapterBase.Exists(dts) Then Throw dts.Exception(
					"only call DataSet.Persist() **before** its Attachement is set")
			For Each tb As DataTable In dts.Tables
				With tb.Attach
					If .Persistance <> Persistance.Guess Then Throw .Exception("Persistance already set")
					.Persistance = If(value = tables.Contains(tb), Persistance.Yes, Persistance.No)
				End With
			Next
		End Sub
		''' <summary>mark all Tables as to persist. All other tables will behave contrary. This call isn't necessary, but optimizes startup</summary>
		<Extension()>
		Public Sub PersistAll(dts As DataSet)
			dts.Persist(False)
		End Sub
		''' <summary>
		''' tables must be an ascending line of tables - each tables CurrentChanged will notify the as next in the list specified parent.
		''' </summary>
		''' <exception cref="System.InvalidOperationException">no Datarelation found between two tables</exception>
		<Extension()>
		Public Sub SetParentNotifications(tb0 As DataTable, ParamArray tables As DataTable())
			tb0.Attach.SetParentNotifications(tables)
			'System.Data.DataTableAttach.SetParentNotifications(tables)
		End Sub
		''' <summary> get Current of DataTable </summary>
		<Extension()>
		Public Function Current(Of T As DataRow)(tb As TypedTableBase(Of T)) As T
			Return DirectCast(tb.Attach.Current, T)
		End Function
		''' <summary> update Current-Tracking of DataTable </summary>
		<Extension()>
		Public Sub Current(Of T As DataRow)(tb As TypedTableBase(Of T), value As T)
			tb.Attach.Current = value
		End Sub
		''' <summary> registers BindingSource to be updated from Current-Tracking of DataTable </summary>
		<Extension()>
		Public Function RegisterCurrency(bs As BindingSource) As DataTableAttach
			Dim cm = bs.DataTable.Attach
			cm.RegisterBindingSource(bs)
			Return cm
		End Function
		''' <summary> registers bindingSources to be updated from Current-Tracking of DataTable </summary>
		<Extension()>
		Public Sub RegisterCurrency(dts As DataSet, ParamArray bindingSources As BindingSource())
			For Each bs In bindingSources
				bs.DataTable.Attach.RegisterBindingSource(bs)
			Next
		End Sub
		''' <summary> returns a detached Datarow (-Template), containing Defaultvalues </summary>
		<Extension()>
		Public Function GetDefaultValues(Of T As DataRow)(tb As TypedTableBase(Of T)) As T
			Return DirectCast(tb.Attach.DefaultValues, T)
		End Function



	End Module

End Namespace
