﻿Imports System.Reflection
Imports System.Collections

Namespace Global.System.Windows.Forms

	Public Class ComboboxCompletionPreview : Inherits CompletionPreviewBase
		'Usage: set one or many Comboboxes, optionally set Base.WindForward-Property

		Private _Combos As IEnumerable(Of Control)

		Public Sub SetCombos(ParamArray combos As ComboBox())
			_Combos = combos
			For Each cmb In combos
				AddHandler cmb.Enter, AddressOf Combo_Enter
				cmb.AutoCompleteMode = AutoCompleteMode.SuggestAppend
				cmb.AutoCompleteSource = AutoCompleteSource.ListItems
			Next
		End Sub

		Private Sub Combo_Enter(sender As Object, e As EventArgs)
			Dim cmb = DirectCast(sender, ComboBox)
			SetControl(cmb)
			Parser.RawWords = cmb.Items.Cast(Of Object).Select(AddressOf cmb.GetItemText).Distinct().ToArray
			Array.Sort(Parser.RawWords, StringComparer.OrdinalIgnoreCase)
			AddHandler cmb.Leave, AddressOf Combo_Leave
		End Sub

		Private Sub Combo_Leave(sender As Object, e As EventArgs)
			RemoveHandler DirectCast(sender, ComboBox).Leave, AddressOf Combo_Leave
			Parser.RawWords = Nothing
		End Sub

	End Class

	Public Class TextboxCompletionPreview : Inherits CompletionPreviewBase
		'Usage: set Textbox-Property (and optionally Base.WindForward-Property), give Data to it and handle InputDone-Event

		Public Event InputDone As EventHandler
		Private WithEvents _TextBox As TextBox

		Public Sub ReadData(texts As IEnumerable(Of String))
			Dim txts = texts.ToArray
			Array.Sort(txts, StringComparer.OrdinalIgnoreCase)
			Dim autoSource = _TextBox.AutoCompleteCustomSource
			autoSource.Clear()
			autoSource.AddRange(txts)
			Dim bf = BindingFlags.Instance Or BindingFlags.NonPublic
			Dim inf = GetType(AutoCompleteStringCollection).GetField("data", bf)
			Dim data = inf.GetValue(autoSource)
			inf = GetType(ArrayList).GetField("_items", bf)
			Parser.RawWords.Be(inf.GetValue(data))
		End Sub

		Public Property Textbox As TextBox
			Get
				Return _TextBox
			End Get
			Set(value As TextBox)
				If _TextBox Is value Then Return
				If _TextBox IsNot Nothing Then Throw New Exception("Textbox already set!")
				value.AutoCompleteMode = AutoCompleteMode.SuggestAppend
				value.AutoCompleteSource = AutoCompleteSource.CustomSource
				value.AutoCompleteCustomSource.Clear()
				SetControl(value)
				_TextBox = value
			End Set
		End Property

		Private Sub _TextBox_Leave(sender As Object, e As EventArgs) Handles _TextBox.Leave
			RaiseEvent InputDone(Me, EventArgs.Empty)
		End Sub
	End Class
End Namespace