Imports System.Drawing
Imports System.ComponentModel

Namespace Global.System.Windows.Forms

	''' <summary>
	''' Dieses Controls Beschriftung läßt sich durch Veränderung seiner
	''' CaptionStringFormat-Property manipulieren
	''' </summary>
	<DebuggerDisplay("{txt}")>
	<DesignerCategory("Code")>
	Public Class SFLabel : Inherits Control

		Private ReadOnly _CaptionStringFormat As StringFormat = New StringFormat
		Private _ForeBrush As New SolidBrush(Color.Black)
		Private _BackBrush As New SolidBrush(Color.White)

		Public Sub New()
			Me.SetStyle(ControlStyles.ResizeRedraw, True)
			MyBase.BackColor = Color.Gainsboro
		End Sub

		Public Overrides Property ForeColor As Drawing.Color
			Get
				Return _ForeBrush.Color
			End Get
			Set(value As Color)
				_ForeBrush.Color = value
			End Set
		End Property

		Private Function txt() As String
			Return Text
		End Function

		Protected Overloads Overrides Sub Dispose(disposing As Boolean)
			If Not MyBase.IsDisposed Then
				For Each D As IDisposable In New IDisposable() {_ForeBrush, _BackBrush, _CaptionStringFormat}
					D.Dispose()
				Next
			End If
			MyBase.Dispose(disposing)
		End Sub

		'Einstellung: Im Designer den Knoten der Property "CaptionStringFormat" erweitern, und Einstellungen vornehmen
		'Es kann das vertikale und horizontale Alignment bestimmt werden, ob der Text umgebrochen wird (FormatFlags), welche Zeichen gekürzt werden, wenn der Text zu lang ist (Trimming)
		'Die Änderungen sind sofort sichtbar - kein Kompilieren erforderlich
		<TypeConverter(GetType(ExpandableObjectConverter))>
		<Browsable(True)>
		<Category("Darstellung")>
		<Description("Einige interessante Einstellmöglichkeiten der Schrift-Darstellung")>
		<DesignerSerializationVisibility(DesignerSerializationVisibility.Content)>
		Public ReadOnly Property CaptionStringFormat() As StringFormat
			Get
				If Me.DesignMode Then Invalidate()
				Return _CaptionStringFormat
			End Get
		End Property

		Protected Overrides Sub OnTextChanged(e As System.EventArgs)
			MyBase.OnTextChanged(e)
			Invalidate()
		End Sub


		Private _BorderStyle As Border3DStyle = Border3DStyle.Adjust

		<DefaultValue(GetType(Border3DStyle), "Adjust")>
		<Category("Darstellung")>
		Public Property BorderStyle() As Border3DStyle
			Get
				Return _BorderStyle
			End Get
			Set(NewValue As Border3DStyle)
				If Object.Equals(NewValue, Me.BorderStyle) Then Return
				_BorderStyle = NewValue
			End Set
		End Property

		Protected Overrides Sub OnPaint(e As PaintEventArgs)
			_ForeBrush.Color.Assign(MyBase.ForeColor)
			_BackBrush.Color.Assign(MyBase.BackColor)
			e.Graphics.FillRectangle(_BackBrush, ClientRectangle)
			e.Graphics.DrawString(MyBase.Text, MyBase.Font, _ForeBrush, ClientRectangle, _CaptionStringFormat)
			ControlPaint.DrawBorder3D(e.Graphics, ClientRectangle, _BorderStyle)
			MyBase.OnPaint(e)
		End Sub

	End Class
End Namespace