﻿Imports System
Imports System.Drawing
Imports System.ComponentModel

Namespace Global.System.Windows.Forms

	' each childControl can be resized (by default)
	' There are several options to modify the layout-behaviour:
	' TopDown As Boolean sets the FlowDirection to TopDown or LeftToRight
	' If WrapContents=True FlowLayout layouts the childControls in one or more lines/columns. In
	'°FlowDirection you can resize each child, in notFlowDirection you can resize the heigth/width of
	'°those lines/columns. 
	' If WrapContents=False, ChildControls become resized to fill all space.
	' If so, you can define a FillControl, which will fill the remaining space, and shrink, when you
	'°enlarge another Control. This behaviour is like SplitContainer, but with more Panels
	<Docking(DockingBehavior.Ask)>
	<ProvideProperty("IsFillControl", GetType(Control))>
	<DesignerCategory("Code")>
	Public Class SizerPanel : Inherits FlowLayoutPanel

		Private Shared _ScrollbarSize As Integer = SystemInformation.HorizontalScrollBarHeight   'same value
		'°                                                                                in both directions

		Private _Sizeable As Control                            'is set in MouseMove, before Dragging starts
		Private _FillCtl As Control
		Private _Anchor As Integer = Integer.MinValue            'is set in MouseDown, to indicate DragStart
		Private _MarkupPen As New Pen(Color.Red, 2) With {.Alignment = Drawing2D.PenAlignment.Inset}
		Private _MarkupPadding As New Padding(CInt(_MarkupPen.Width))     'marking the FillControl takes its
		'°                                  bounds, enlarges it with this Padding, and colorizes the Padding

		Private _SizeableIndex As Integer                       'index of _Sizeable in the ControlCollection
		Private _FillCtlIndex As Integer = Integer.MaxValue     'MaxValue enshures a behavior like _Sizeable
		'°                                                       would be before _FillCtl, if no FillCtl set
		Private Shared _Cursors As Cursor() = {Cursors.VSplit, Cursors.HSplit}
		'Private Shared ReadOnly OldSize As New ComponentProperty(Of Control, Size) _
		'(Function(ctl) SizeMinValue)
		'Private _Tabs As List(Of Integer) = Nothing                     'widths/heights of the lines/columns
		Private _Initing As Boolean = True
		Private _Bounds As BoundsEx
		Private _Tabs As New List(Of Integer) ' either horizontal or vertical Tabs
		Private _OldSize As New Dictionary(Of Control, Vector)

		Public Sub New()
			MyBase.AutoScroll = True
			_Bounds = BoundsEx.CreateEntry(Me)
			_Bounds.ReInit()
		End Sub

#Region "Publix"



		<Category("Layout")>
		<DefaultValue(True)>
		<Description("True: Align controls in several lines." & Lf & "False: enlarge heights to fill out the containing Sizerpanel")>
		Public Shadows Property WrapContents() As Boolean
			Get
				Return MyBase.WrapContents
			End Get
			Set(value As Boolean)
				MyBase.WrapContents = value
			End Set
		End Property

		<Category("Layout")>
		<DefaultValue(True)>
		<Description("If False: AutoSize-Behavior")>
		Public Overrides Property AutoScroll() As Boolean
			Get
				Return MyBase.AutoScroll
			End Get
			Set(value As Boolean)
				MyBase.AutoScroll = value
			End Set
		End Property

		<Category("Layout")>
		<DefaultValue(False)>
		<DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)>
		<Description("True: FlowDirection TopDown" & Lf & "False: FlowDirection LeftToRight")>
		Public Property TopDown() As Boolean
			'reduces the number of the FlowDirection-Options to only two
			Get
				Return MyBase.FlowDirection = Forms.FlowDirection.TopDown
			End Get
			Set(NewValue As Boolean)
				If NewValue = (MyBase.FlowDirection = Forms.FlowDirection.TopDown) Then Return
				MyBase.FlowDirection = If(NewValue, FlowDirection.TopDown, FlowDirection.LeftToRight)
			End Set
		End Property

		<Category("Appearance")>
		Public Property MarkupColor() As Color
			' when Wrapmode=False there may be set the FillControl, which acts like a control with
			'°DockStyle.Fill: it shrinks, if the choosen control grows, and grows, if the choosen one
			'°shrinks. So there are always **two** Controls sizing, when SplitBar is used. MarkupColor is
			'°the color used to indicate them.
			Get
				Return _MarkupPen.Color
			End Get
			Set(NewValue As Color)
				If _MarkupPen.Color.ToArgb = NewValue.ToArgb Then Return
				_MarkupPen.Color = NewValue
				InvalidateMarkup()
			End Set
		End Property

#Region "Publix.ProvideProperties"

		<Category("Layout")>
		<Description("when Wrapmode=False the FillControl is layouted with DockStyle.Fill-Behaviour")>
		<DefaultValue(False), DisplayName("IsFillControl")>
		Public Function GetIsFillControl(control As Control) As Boolean
			Return control Is _FillCtl
		End Function

		<DisplayName("IsFillControl")>
		Public Sub SetIsFillControl(control As Control, value As Boolean)
			If (control Is _FillCtl) = value Then Return
			Dim oldValue = _FillCtl.NotNull
			InvalidateMarkup()
			_FillCtl = If(value, control, Nothing)
			InvalidateMarkup()
			If value <> oldValue AndAlso Not _Initing Then Me.PerformLayout(Me, "IsFillControl")
		End Sub

#End Region 'Publix.ProvideProperties

#Region "Publix.HideProperties"
		<Browsable(False)>
		<EditorBrowsable(EditorBrowsableState.Never)>
		Public Shadows Property FlowDirection() As FlowDirection
			Get
				Return MyBase.FlowDirection
			End Get
			Set(value As FlowDirection)
				MyBase.FlowDirection = value
			End Set
		End Property
		<Browsable(False)>
		<EditorBrowsable(EditorBrowsableState.Never)>
		Public Shadows Property AutoScrollMinSize() As Size
			Get
				Return MyBase.AutoScrollMinSize
			End Get
			Set(value As Size)
				MyBase.AutoScrollMinSize = value
			End Set
		End Property
		<Browsable(False)>
		<EditorBrowsable(EditorBrowsableState.Never)>
		Public Overrides Property AutoSize() As Boolean
			Get
				Return False
			End Get
			Set(value As Boolean)
			End Set
		End Property
		<Browsable(False)>
		<EditorBrowsable(EditorBrowsableState.Never)>
		Public Overrides Property AutoSizeMode() As AutoSizeMode
			Get
				Return MyBase.AutoSizeMode
			End Get
			Set(value As AutoSizeMode)
				MyBase.AutoSizeMode = value
			End Set
		End Property
		<Browsable(False)>
		<EditorBrowsable(EditorBrowsableState.Never)>
		Public Overrides Property RightToLeft() As RightToLeft
			Get
				Return MyBase.RightToLeft
			End Get
			Set(value As RightToLeft)
				MyBase.RightToLeft = value
			End Set
		End Property
		<Browsable(False)>
		<EditorBrowsable(EditorBrowsableState.Never)>
		Public Overrides Property Cursor() As Cursor
			Get
				Return MyBase.Cursor
			End Get
			Set(value As Cursor)
				MyBase.Cursor = value
			End Set
		End Property
		<Browsable(False)>
		Public Shadows Property UseWaitCursor() As Boolean
			Get
				Return MyBase.UseWaitCursor
			End Get
			Set(value As Boolean)
				MyBase.UseWaitCursor = value
			End Set
		End Property
#End Region 'Publix.HideProperties

#End Region 'Publix

#Region "Overrides"

		Protected Overrides Sub OnLayout(e As LayoutEventArgs)
			Dim affected = e.AffectedControl
			If _Initing Then
				' suppress Layout until first Me.Bounds - LayoutRequest. So
				'°user-settings do not override each other when loading from Form.Designer.vb
				If Parent.Null OrElse affected IsNot Me Then Return
				If NotifyPropertyChanged.IsInDesignMode AndAlso e.AffectedProperty <> "Bounds" Then Return
				_OldSize(Me) = Size
				_Initing = False
			End If
			'DesignDebugger.Msg(If(affected.Null, "##", affected.Name), e.AffectedProperty)
			Dim bnds = New List(Of BoundsEx)(Controls.Count)
			Dim bndSelected As BoundsEx = Nothing
			Dim bndFilling As BoundsEx = Nothing
			Dim notFlow = MyBase.FlowDirection Xor 1
			For Each ctl As Control In Controls
				If Not ctl.Visible Then Continue For
				Dim bnd = ctl.BoundsX
				bnd.ReInit()
				If ctl Is affected AndAlso e.AffectedProperty = "Bounds" Then
					Select Case bnd.Size.DifferenceDirection(_OldSize.GetOrCreate(ctl))
						Case notFlow, 2
							bndSelected = bnd
						Case -1
							Return ' suppress Layout if Size didn't change
					End Select
					_OldSize(bnd.Ctl) = bnd.Size
				End If
				If ctl Is _FillCtl Then bndFilling = bnd
				bnds.Add(bnd)
			Next
			If bnds.Count = 0 Then Return
			_Bounds.ReInit()
			If bndFilling.Null Then _FillCtl = Nothing
			If affected Is Me Then
				Select Case e.AffectedProperty
					Case "IsFillControl"
						Invalidate()
					Case "Bounds"
						If bndFilling.Null OrElse Not MyBase.WrapContents Then Exit Select
						Dim dlt = _Bounds.Size(notFlow) - _OldSize.GetOrCreate(Me)(notFlow)
						If dlt = 0 Then Exit Select
						bndFilling.Size(notFlow) += dlt
						bndSelected = bndFilling
					Case "FlowDirection", "WrapContents"
						For Each bnd In bnds
							bnd.Size = BoundsEx.ControlDefault
							bnd.ApplyBounds(True)
						Next
						_Tabs.Clear()
						Invalidate()
				End Select
			End If
			If MyBase.WrapContents Then
				LayoutWrapping(bnds, bndSelected)
			Else
				LayoutNoWrap(bnds)
			End If
			If MyBase.AutoScroll Then MyBase.AdjustFormScrollbars(True)
			MyBase.OnLayout(e)
			_OldSize(Me) = Size
		End Sub

		Protected Overrides Sub Dispose(disposing As Boolean)
			If disposing AndAlso _MarkupPen IsNot Nothing Then _MarkupPen.Dispose()
			MyBase.Dispose(disposing)
		End Sub

		Protected Overrides Sub OnPaint(e As PaintEventArgs)
			MyBase.OnPaint(e)
			'paint enclosing Rectangles around _FillCtl and _Sizeable
			If _FillCtl.Null Then Return
			If NotifyPropertyChanged.IsInDesignMode Then
				e.Graphics.DrawRectangle(_MarkupPen, _FillCtl.Bounds.Add(_MarkupPadding))
			ElseIf Not MyBase.WrapContents AndAlso _Anchor <> Integer.MinValue AndAlso Controls.Count > 2 Then
				e.Graphics.DrawRectangle(_MarkupPen, _FillCtl.Bounds.Add(_MarkupPadding))
				e.Graphics.DrawRectangle(_MarkupPen, _Sizeable.Bounds.Add(_MarkupPadding))
			End If
		End Sub

		Protected Overrides Sub OnMouseLeave(e As EventArgs)
			MyBase.OnMouseLeave(e)
			Cursor = Cursors.Default
		End Sub

		Protected Overrides Sub OnMouseUp(e As MouseEventArgs)
			MyBase.OnMouseUp(e)
			InvalidateMarkup()
			_Anchor = Integer.MinValue
		End Sub

		Protected Overrides Sub OnMouseDown(e As MouseEventArgs)
			MyBase.OnMouseDown(e)
			' _Anchor is the difference between the mouseposition and the height of the sizing control. It
			'°can have 4 meanings: X/Y - Dimension, and before/behind the FillControl. If the sizing control
			'°is before the FillingControl it grows when mouse.Y grows. But if it is behind the
			'°FillingControl it shrinks when mouse.Y grows.
			Dim i = _Cursors.IndexOf(Cursor)                             '0: horizontal sizement, 1: vertical
			If i < 0 Then Return
			Dim sizeAmount = _Sizeable.Size.ToVector(i)
			_Anchor = e.Location.ToVector(i) + If(_SizeableIndex > _FillCtlIndex, sizeAmount, -sizeAmount)
			If Not MyBase.WrapContents Then InvalidateMarkup()
		End Sub

		Protected Overrides Sub OnMouseMove(e As MouseEventArgs)
			MyBase.OnMouseMove(e)
			Select Case e.Button
				Case MouseButtons.None
					'check out, wether there is a control to size - then set Cursor to HSplit/VSplit
					_FillCtlIndex = Integer.MaxValue
					Dim iFlow = FlowDirection
					If MyBase.WrapContents Then
						For i = 0 To 1
							If TrySetSizeable(e.Location, i) Then
								Cursor = _Cursors(i)
								Return
							End If
						Next
					ElseIf TrySetSizeableNoWrap(e.Location, iFlow) Then
						Cursor = _Cursors(iFlow)
						Return
					End If
					Cursor = Cursors.Default
				Case MouseButtons.Left
					' execute sizing depending on the cursor
					Dim i = _Cursors.IndexOf(Cursor)
					If i < 0 Then Return
					InvalidateMarkup()                                                        'clear old markup
					Dim bnd = _Sizeable.BoundsX
					Dim anchorAmount = e.Location.ToVector(i) - _Anchor
					Dim dlt = If(_SizeableIndex > _FillCtlIndex, -anchorAmount, anchorAmount) - bnd.Size(i)
					If _FillCtlIndex <> Integer.MaxValue Then
						With _FillCtl.BoundsX
							' dlt also can be limited by FillControls MinimumSize
							dlt.Minimize(.Size(i) - .Min(i))
						End With
					End If
					bnd.Size(i) += dlt
					bnd.ApplyBounds(True)     'limitation by _Sizeable itself is enshured by ApplyBounds
				Case Else
			End Select
		End Sub

		Protected Overrides Sub OnControlRemoved(e As ControlEventArgs)
			MyBase.OnControlRemoved(e)
			_OldSize.Remove(e.Control)
		End Sub


		Public Overrides ReadOnly Property LayoutEngine() As Layout.LayoutEngine
			Get
				Return NoLayout.Instance
			End Get
		End Property

#End Region 'Overrides

#Region "Privates"

		Private Sub InvalidateMarkup()
			'request Painting of enclosing Rectangles around _FillCtl and _Sizeable
			If MyBase.WrapContents OrElse _FillCtl.Null Then Return
			Invalidate(_FillCtl.Bounds.Add(_MarkupPadding).Add(Size1))
			If _Sizeable.Null Then Return
			Invalidate(_Sizeable.Bounds.Add(_MarkupPadding).Add(Size1))
		End Sub

		Private Function TrySetSizeable(pt As Point, iFlow As Integer) As Boolean
			'try set _Sizeable to a Control nearby (top or left) the point
			Dim checkpoints = {pt.Add(-15, 0), pt.Add(0, -15)}
			_Sizeable = Me.GetChildAtPoint(checkpoints(iFlow), GetChildAtPointSkip.Invisible)
			Return _Sizeable.NotNull AndAlso _Sizeable.BoundsX.CanSize(iFlow)
		End Function

		Private Function TrySetSizeableNoWrap(pt As Point, iFlow As Integer) As Boolean
			'try set _Sizeable to a Control nearby (top or left) the point
			TrySetSizeableNoWrap = TrySetSizeable(pt, iFlow)
			If _FillCtl.Null Then Exit Function
			If TrySetSizeableNoWrap Then
				'when WrapContent=False and _FillCtl is set, figuring out the _Sizeable gets tricky
				_SizeableIndex = Controls.GetChildIndex(_Sizeable)
				_FillCtlIndex = Controls.GetChildIndex(_FillCtl)
				If _SizeableIndex = _FillCtlIndex Then
					'if FillCtl hit, search a sizeable Control after it
					For Me._SizeableIndex = _SizeableIndex + 1 To Controls.Count - 1
						If Controls(_SizeableIndex).BoundsX.CanSize(iFlow) Then
							_Sizeable = Controls(_SizeableIndex)
							Return True
						End If
					Next
				ElseIf _SizeableIndex > _FillCtlIndex Then
					' if a Control after FillCtl hit, take the Control below, because sizing now moves the top-
					'°edge
					_SizeableIndex += 1
					If Controls.Count = _SizeableIndex Then Return False
					_Sizeable = Me.Controls(_SizeableIndex)
					Return _Sizeable.BoundsX.CanSize(iFlow)
				End If
			End If
		End Function

		Private Sub LayoutNoWrap(bnds As List(Of BoundsEx))
			' fill out SizerPanel in notFlowDirection, try size _FillCtl, that SizerPanel is filled out,
			'°without Scrollbars
			Dim iFlow = MyBase.FlowDirection
			Dim notFlow = iFlow Xor 1
			Dim available = (ClientSize - Padding.Size).ToVector
			Dim ctlSum As Integer = 0
			For Each bnd In bnds
				With bnd
					ctlSum += .Size(iFlow)
					.Size(notFlow) = available(notFlow) - (.MarginSize(notFlow))     'fill out
					'°                                                           SizerPanel in notFlowDirection
					available(iFlow) -= .MarginSize(iFlow)            'respect Margins influence to sizeability
				End With
			Next
			If _FillCtl.NotNull Then                        'let _FillCtl fill out the rest in FlowDirection 
				With _FillCtl.BoundsX
					.Size(iFlow) += available(iFlow) - ctlSum
					' now ctlSum should be equal to available(iFlow). But if .Size is set smaller than .Min,
					'°it will be rejected, so that ctlSum gets bigger than available. This is important for
					'°anticipating scrollbar-appearance
					ctlSum = available(iFlow) + .Min(iFlow) - .Size(iFlow)
				End With
			End If
			If MyBase.AutoScroll AndAlso ctlSum > available(iFlow) _
					AndAlso Not {HScroll, VScroll}(iFlow) Then
				'anticipate, that a new scrollbar will appear
				For Each bnd In bnds
					bnd.Size(notFlow) -= _ScrollbarSize
				Next
			End If
			Dim curr = Padding.Vector1 + MyBase.AutoScrollPosition
			For Each bnd In bnds
				bnd.TopLeft = curr
				curr(iFlow) += bnd.Fullsize(iFlow)
				bnd.ApplyBounds(True)
			Next
		End Sub

		Private Sub LayoutWrapping(bnds As List(Of BoundsEx), bndSelected As BoundsEx)
			Dim iFlow = MyBase.FlowDirection
			Dim notFlow = iFlow Xor 1
			Dim curr As Vector = Padding.Vector1 + AutoScrollPosition
			Dim available As Vector = ClientSize - Padding.Vector2
			Dim ground = curr(iFlow)
			Dim iLineStart As Integer = 0
			Dim newTab As Integer = 0
			Dim tabPos As Integer = 0
			For iBnd = 0 To bnds.Count - 1
				Dim bnd = bnds(iBnd)
				If curr(iFlow) + bnd.Fullsize(iFlow) > available(iFlow) Then
					If tabPos = _Tabs.Count Then _Tabs.Add(newTab)
					AssignLineHeight(iLineStart, iBnd - 1, bnds, tabPos)
					curr(iFlow) = ground
					curr(notFlow) += _Tabs(tabPos)
					tabPos += 1
					iLineStart = iBnd
					newTab = 0
				End If
				If bnd Is bndSelected Then
					If tabPos = _Tabs.Count Then
						_Tabs.Add(bnd.Fullsize(notFlow))
					Else
						_Tabs(tabPos) = bnd.Fullsize(notFlow)
					End If
				ElseIf tabPos = _Tabs.Count Then
					newTab.Maximize(bnd.Fullsize(notFlow))
				End If
				bnd.TopLeft = curr
				curr(iFlow) += bnd.Fullsize(iFlow)
			Next
			If tabPos = _Tabs.Count Then _Tabs.Add(newTab)
			AssignLineHeight(iLineStart, bnds.Count - 1, bnds, tabPos)
			For Each bnd In bnds
				bnd.ApplyBounds(False)
			Next
			If Not AutoScroll Then
				newTab = _Tabs(tabPos) + curr(notFlow)
				If newTab <> available(notFlow) Then
					available = ClientSize
					available(notFlow) = newTab + Padding.Vector2(notFlow)
					ClientSize = available
				End If
			End If
			tabPos += 1
			_Tabs.RemoveRange(tabPos, _Tabs.Count - tabPos)
		End Sub

		Private Sub AssignLineHeight(iLine As Integer, ibnd As Integer,
				 bnds As List(Of BoundsEx), tabPos As Integer)
			Dim notFlow = MyBase.FlowDirection Xor 1
			Dim tab = _Tabs(tabPos)
			For i = iLine To ibnd
				With bnds(i)
					Dim pad = .TrySetFullSize(tab, notFlow)      'to set Fullsize can be rejected by CanSize or
					'°                                                              limited by Mini/MaximumSize
					If pad > 1 Then .Location(notFlow) += pad \ 2 ' if so, align centered
				End With
			Next
		End Sub

#End Region 'Privates

	End Class

End Namespace
