﻿Imports System.Linq
Imports System.Collections
Imports System.Collections.Generic
Imports System.Data
Imports System.Diagnostics
Imports System.Runtime.CompilerServices
Imports System.ComponentModel
Imports System.Reflection

Namespace Global.System.Windows.Forms

	''' <summary>establishes Synchronisization (two-way by default) of parent-bs.Position with child-bs.foreignkey.
	''' note: requires a DataRelation (any direction) to the data of the other BindingSource. </summary>
	Public Class BindingSourceBinder

		''' <summary>suppresses entrying a changed Parent-Records PK as FK into the currend Child-Record. Note! this can be seen as inconsistent Data-Presentation!</summary>
		Public SuppressSetParentToChild As Boolean = False
		Private rl As DataRelation
		Private WithEvents bsParent As BindingSource
		Private WithEvents bsChild As BindingSource
		Private _OldParent As DataRow
		Private _SetParentRowCallback As Action = Sub()
																	Dim tpl = TryGet_ChildParentRow_Tuple()
																	tpl?.Item1.SetParentRow(tpl.Item2, rl)
																End Sub

		Private Function TryGet_ChildParentRow_Tuple() As Tuple(Of DataRow, DataRow)
			Dim rwChild = bsChild.At
			If rwChild.Null Then Return Nothing
			Dim rwParent = bsParent.At
			If rwParent?.RowState = DataRowState.Detached OrElse rwChild.GetParentRow(rl) Is rwParent Then Return Nothing
			Return Tuple.Create(rwChild, rwParent)
		End Function

		Private Sub bsParent_CurrentChanged(sender As Object, e As EventArgs) Handles bsParent.CurrentChanged
			If SuppressSetParentToChild OrElse TryGet_ChildParentRow_Tuple() Is Nothing Then Return
			BusyDelay.SetCallback(_SetParentRowCallback)
		End Sub

		Private Sub bsChild_CurrentChanged(sender As Object, e As EventArgs)
			Dim rwChild = bsChild.At
			If rwChild.Null Then Return
			Dim prev = SuppressSetParentToChild
			SuppressSetParentToChild = True
			_OldParent = rwChild.GetParentRow(rl)
			bsParent.MoveToRow(_OldParent)
			SuppressSetParentToChild = prev
		End Sub

		Public Sub New(first As BindingSource, second As BindingSource)
			Dim dic = {first, second}.ToDictionary(Function(src) src.DataTable)
			Dim rl = dic.First.Key.DataSet.Relations.Cast(Of DataRelation).FirstOrDefault(
					 Function(r) dic.ContainsKey(r.ChildTable) AndAlso dic.ContainsKey(r.ParentTable))
			If rl.Null Then Throw Me.Exception("there is no DataRelation between ", first.DataTable.TableName, " and ", second.DataTable.TableName)
			Me.bsParent = dic(rl.ParentTable)
			Me.bsChild = dic(rl.ChildTable)
			Me.rl = rl
			AddHandler bsChild.CurrentChanged, AddressOf bsChild_CurrentChanged
			bsChild_CurrentChanged(second, Nothing)
		End Sub

		Private Sub bsChild_ListChanged(sender As Object, e As ListChangedEventArgs) Handles bsChild.ListChanged
			If _OldParent IsNot bsParent.At Then bsChild_CurrentChanged(bsChild, Nothing)
		End Sub
		Private Sub bsParent_ListChanged(sender As Object, e As ListChangedEventArgs) Handles bsParent.ListChanged
			If bsParent.Sort <> "" <> bsParent.IsSorted Then Return
			BusyDelay.RemoveCallback(_SetParentRowCallback)
		End Sub
	End Class

End Namespace
#If False Then
Imports System.Linq
Imports System.Collections
Imports System.Collections.Generic
Imports System.Data
Imports System.Diagnostics
Imports System.Runtime.CompilerServices
Imports System.ComponentModel
Imports System.Reflection

Namespace Global.System.Windows.Forms

   ''' <summary>establishes Synchronisization (two-way) of parent-bs.Position with child-bs.foreignkey.
   ''' note: requires a DataRelation (any direction) to the data of the other BindingSource. </summary>
   Public Class BindingSourceBinder
      Private rl As DataRelation
      Private WithEvents bsParent As BindingSource
      Private WithEvents bsChild As BindingSource
      Private _OldParent As DataRow
      Private _DownCallback As Action = Sub()
                                           Dim rwChild = bsChild.At
                                           If rwChild.Null Then Return
                                           Dim rwParent = bsParent.At
                                           If rwParent.Null OrElse rwChild.GetParentRow(rl) Is rwParent Then Return
                                           'rwChild.SetParentRow(rwParent, rwChild.Table.ParentRelations(rl.RelationName))
                                           rwChild.SetParentRow(rwParent, rl)
                                           'rwChild(rl.ChildColumns(0).ColumnName) = rwParent(rl.ParentColumns(0))
                                        End Sub
      Private Sub bsParent_CurrentChanged(sender As Object, e As EventArgs)
         Dim rwChild = bsChild.At
         If rwChild.Null Then Return
         Dim rwParent = bsParent.At
         If rwParent.Null OrElse rwParent.RowState = DataRowState.Detached OrElse rwChild.GetParentRow(rl) Is rwParent Then Return
         BusyDelay.SetCallback(_DownCallback)
      End Sub
      Private Sub bsChild_CurrentChanged(sender As Object, e As EventArgs)
         Dim rwChild = bsChild.At
         If rwChild.Null Then Return
         RemoveHandler bsParent.CurrentChanged, AddressOf bsParent_CurrentChanged
         _OldParent = rwChild.GetParentRow(rl)
         bsParent.MoveToRow(_OldParent)
         AddHandler bsParent.CurrentChanged, AddressOf bsParent_CurrentChanged
      End Sub

      Public Sub New(first As BindingSource, second As BindingSource)
         Dim dic = {first, second}.ToDictionary(Function(src) src.DataTable)
         Dim rl = dic.First.Key.DataSet.Relations.Cast(Of DataRelation).FirstOrDefault( _
            Function(r) dic.ContainsKey(r.ChildTable) AndAlso dic.ContainsKey(r.ParentTable))
         If rl.Null Then Throw Me.Exception("there is no DataRelation between ", first.DataTable.TableName, " and ", second.DataTable.TableName)
         Me.bsParent = dic(rl.ParentTable)
         Me.bsChild = dic(rl.ChildTable)
         Me.rl = rl
         AddHandler bsChild.CurrentChanged, AddressOf bsChild_CurrentChanged
         bsChild_CurrentChanged(second, Nothing)
      End Sub

      Private Sub bsChild_ListChanged(sender As Object, e As ListChangedEventArgs) Handles bsChild.ListChanged
         If _OldParent IsNot bsParent.At Then bsChild_CurrentChanged(bsChild, Nothing)
      End Sub
      Private Sub bsParent_ListChanged(sender As Object, e As ListChangedEventArgs) Handles bsParent.ListChanged
         If bsParent.Sort <> "" <> bsParent.IsSorted Then Return
         BusyDelay.RemoveCallback(_DownCallback)
      End Sub
   End Class

End Namespace


#End If