﻿
Imports System.ComponentModel
Imports System.Runtime.CompilerServices

<EditorBrowsable(EditorBrowsableState.Never)>
Public Module DataGridViewX
	<Extension>
	Public Function Attach(dgv As DataGridView) As DataGridViewAttach
		Return DataGridViewAttach.Get(dgv)
	End Function
	''' <summary> enumeriert die DataRowViews, bei denen die CheckboxCell gecheckt ist </summary>
	<Extension()>
	Public Iterator Function CheckedDrvs(col As DataGridViewCheckBoxColumn) As IEnumerable(Of DataRowView)
		Dim dgv = col.DataGridView
		Dim colIndx = col.Index
		For i = 0 To dgv.RowCount - 1 - dgv.AllowUserToAddRows.AsIndex
			Dim rwDgv = dgv.Rows(i)
			If CBool(rwDgv.Cells(colIndx).EditedFormattedValue) Then Yield DirectCast(rwDgv.DataBoundItem, DataRowView)
		Next
	End Function
	''' <summary> enumeriert die DataRows, bei denen die CheckboxCell gecheckt ist </summary>
	<Extension()>
	Public Function CheckedDrvs(Of T As DataRow)(col As DataGridViewCheckBoxColumn) As IEnumerable(Of T)
		Return From drv In col.CheckedDrvs Select DirectCast(drv.Row, T)
	End Function
	<Extension()>
	Public Sub SetValue(col As DataGridViewColumn, rowIndex As Integer, value As Object)
		col.DataGridView(col.Index, rowIndex).Value = value
	End Sub
	''' <summary> returns null if invalid rowIndex </summary>
	<Extension()>
	Public Function DataRow(Of T As DataRow)(grd As DataGridView, rowIndex As Integer) As T
		If rowIndex < 0 Then Return Nothing
		Return grd.Rows(rowIndex).DataRow(Of T)()
	End Function
	<Extension()>
	Public Function DataRow(Of T As DataRow)(dgvRow As DataGridViewRow) As T
		Dim itm = DirectCast(dgvRow.DataBoundItem, DataRowView)
		Return If(itm.Null, Nothing, DirectCast(itm.Row, T))
	End Function
	<Extension()>
	Public Function BoundItem(Of T)(dgvRow As DataGridViewRow) As T
		Return DirectCast(dgvRow.DataBoundItem, T)
	End Function
	<Extension()>
	Public Function DataRows(Of T As DataRow)(dgvRows As IEnumerable(Of DataGridViewRow)) As T()
		Dim drvs = From x In dgvRows Let drv = DirectCast(x.DataBoundItem, DataRowView) Where drv IsNot Nothing Select drv
		Return drvs.ToArray(Function(x) DirectCast(x.Row, T))
	End Function
	<Extension()>
	Public Function BoundItems(Of T)(dgvRows As IEnumerable(Of DataGridViewRow)) As T()
		Return Aggregate x In dgvRows Let itm = DirectCast(x.DataBoundItem, T) Where itm IsNot Nothing Select itm Into ToArray
	End Function
	<Extension()>
	Public Function SelectedDataRows(Of T As DataRow)(dgv As DataGridView) As T()
		Return dgv.SelectedRows.Cast(Of DataGridViewRow).DataRows(Of T)
	End Function
	''' <summary> returns null if invalid rowIndex </summary>
	<Extension()>
	Public Function DataRow(grd As DataGridView, rowIndex As Integer) As DataRow
		If rowIndex < 0 Then Return Nothing
		Dim itm = DirectCast(grd.Rows(rowIndex).DataBoundItem, DataRowView)
		Return If(itm.Null, Nothing, itm.Row)
	End Function

	<Extension()>
	Public Function IsColumnBound(grd As DataGridView, columnIndex As Integer, dataColumn As DataColumn) As Boolean
		If columnIndex < 0 Then Return False
		Return grd.Columns(columnIndex).DataPropertyName = dataColumn.ColumnName
	End Function
	<Extension()>
	Public Function Column(grd As DataGridView, dataColumn As DataColumn) As DataGridViewColumn
		Return grd.ColumnsX(dataColumn).First
	End Function
	<Extension()>
	Public Function ColumnsX(grd As DataGridView, dataColumn As DataColumn) As IEnumerable(Of DataGridViewColumn)
		Return grd.Columns.Cast(Of DataGridViewColumn).Where(Function(col) col.DataPropertyName = dataColumn.ColumnName)
	End Function

End Module
