﻿#Region "FileHeader"
#If False Then
Zustandsfreie "Behaviors" für DatagridView (zustands**behaftete** Behaviors siehe in "DataGridViewAttach.vb")
Die Behaviors hier sind einfach EventHandler, auf den man von einem DGV das entsprechende Event leiten kann, sodass es sich verhält.

• DefaultValueOnDelete_KeyDown(): leitet man ein DGV_KeyDown da drauf, so wird in alle selecteten Zellen ihr Defaultwert eingetragen, wenn der User Delete drückt. Der EditModus muss hierfür nicht betreten werden.

• Grid_DataError(): eine Standard-Meldung für wenn der User Mist in eine DGV-Zelle eingegeben hat.

• Currency_CellValidating(): Zellvalidierung für Decimal-Spalten (bei Währung üblich)

• NegativeRed_CellPainting(): zeichnet in numerischen Spalten negative Zahlen rot

• HideZeroNegativeRed_CellPainting: unterdrückt zusätzlich die Anzeige von 0

#End If '-- Options, Imports
Imports System.Reflection
Imports System.ComponentModel
#End Region 'FileHeader

Namespace Global.System.Windows.Forms

	Public Class DatagridViewBehavior

		Private Shared _GetEvents As FieldInfo = GetType(Component).GetField("events", BindingFlags.NonPublic Or BindingFlags.Instance)
		Private Shared _CurrentChangedKey As Object = GetType(DataGridView) _
			.GetField("EVENT_DATAGRIDVIEWCELLCONTENTCLICK", BindingFlags.NonPublic Or BindingFlags.Static) _
			.GetValue(Nothing) ' that key-name was figgured out that by Disassembler (alternative: Rosylin-website)
		Private Shared Sub InsertCurrentChangedHandler(bs As DataGridView, handler As DataGridViewCellEventHandler)
			Dim events = DirectCast(_GetEvents.GetValue(bs), EventHandlerList)
			Dim prevEvent = events(_CurrentChangedKey)
			Dim newEvent As [Delegate] = handler
			If prevEvent.NotNull Then
				For Each dlg In prevEvent.GetInvocationList
					newEvent = [Delegate].Combine(newEvent, dlg)
				Next
			End If
			events(_CurrentChangedKey) = newEvent
		End Sub
		Private Shared OnCheckbox_CheckEndEdit As DataGridViewCellEventHandler _
			= Sub(s, e)
				  Dim dgv = DirectCast(s, DataGridView)
				  If TypeOf dgv.Columns(e.ColumnIndex) Is DataGridViewCheckBoxColumn Then dgv.EndEdit()
			  End Sub

		''' <summary> enforces immediately accept dgv-Input on (un-)checking on dgv-checkboxColumns </summary>
		Public Shared Sub OnCheckboxCheckEndEdit(dgv As DataGridView)
			InsertCurrentChangedHandler(dgv, OnCheckbox_CheckEndEdit)
		End Sub




		''' <summary> on Key.Delete set all selected Cells to DataColumn.DefaultValues </summary>
		Public Shared Sub DefaultValueOnDelete_KeyDown(sender As Object, e As KeyEventArgs)
			If e.KeyData <> Keys.Delete Then Return
			Dim grd = DirectCast(sender, DataGridView)
			If grd.SelectedRows.Count > 0 Then Return
			e.Handled = True
			For Each c As DataGridViewCell In grd.SelectedCells
				With c.OwningColumn
					If .ReadOnly OrElse "" = .DataPropertyName Then Continue For
					Dim rw = DirectCast(c.OwningRow.DataBoundItem, DataRowView).Row
					Dim dataCol = rw.Table.Columns(.DataPropertyName)
					If dataCol.DefaultValue Is Convert.DBNull AndAlso Not dataCol.AllowDBNull Then Continue For
					rw(dataCol) = dataCol.DefaultValue
				End With
			Next
		End Sub

		Public Shared Sub Grid_DataError(sender As Object, e As DataGridViewDataErrorEventArgs)
			Dim grd = DirectCast(sender, DataGridView)
			Dim s = "Eine gravierende Fehl-Eingabe wurde festgestellt (wie etwa eine Text-Eingabe in ein Zahlenfeld)".And2("\n\nFehler: ", e.Exception.Message)
			MessageBox.Show(grd.FindForm, s, "Eingabe abgelehnt")
			grd.CancelEdit()
		End Sub

		Public Shared Sub Currency_CellValidating(sender As Object, e As DataGridViewCellValidatingEventArgs)
			Dim grd = DirectCast(sender, DataGridView)
			Dim cell = grd(e.ColumnIndex, e.RowIndex)
			If cell.Value Is Nothing OrElse Not cell.ValueType.Equals(GetType(Decimal)) Then Return
			Dim num As Decimal
			e.Cancel = Not Decimal.TryParse(e.FormattedValue.ToString, num) OrElse num Mod 0.01D <> 0
			If e.Cancel Then MessageBox.Show(grd.FindForm, "Zulässig sind nur Zahlen, mit maximal 2 Nachkommastellen", "Eingabefehler")
		End Sub

		Public Shared NegativeRed_CellPainting As DataGridViewCellPaintingEventHandler _
			= Sub(s, e) NumberPainting(DirectCast(s, DataGridView), e, False, True)

		Public Shared HideZeroNegativeRed_CellPainting As DataGridViewCellPaintingEventHandler _
			= Sub(s, e) NumberPainting(DirectCast(s, DataGridView), e, True, True)

		Public Shared HideZero_CellPainting As DataGridViewCellPaintingEventHandler _
			= Sub(s, e) NumberPainting(DirectCast(s, DataGridView), e, True, False)

#Region "Privates"
		Private Shared _ObservedTypes As IList(Of Type) = New Type() {GetType(Integer), GetType(Double), GetType(Decimal)}
		Private Shared _TextFlags As TextFormatFlags = TextFormatFlags.Right Or TextFormatFlags.VerticalCenter
		Private Shared Sub NumberPainting(grd As DataGridView, e As DataGridViewCellPaintingEventArgs, hideZero As Boolean, paintRead As Boolean)
			If e.RowIndex < 0 OrElse e.ColumnIndex < 0 OrElse e.Value Is Nothing Then Return
			If TypeOf grd.Columns(e.ColumnIndex) Is DataGridViewComboBoxColumn Then Return
			If Not _ObservedTypes.Contains(e.Value.GetType) Then Return
			Dim val = CType(e.Value, Decimal)
			If Not hideZero AndAlso val = 0 Then Return
			If val > 0 Then Return
			If val < 0 AndAlso Not paintRead Then Return
			e.Handled = True
			Dim mask = DataGridViewPaintParts.All Xor DataGridViewPaintParts.ContentForeground
			e.Paint(e.CellBounds, e.PaintParts And mask)
			If val = 0 Then Return
			Dim col = grd.Columns(e.ColumnIndex)
			TextRenderer.DrawText(e.Graphics, e.FormattedValue.ToString, e.CellStyle.Font, Shrink(e.CellBounds, e.CellStyle.Padding), Color.Red, _TextFlags)
		End Sub
		Private Shared Function Shrink(rct As Rectangle, pd As System.Windows.Forms.Padding) As Rectangle
			With rct
				Return New Rectangle(.X + pd.Left, .Y + pd.Top, .Width - pd.Horizontal, .Height - pd.Vertical)
			End With
		End Function
#End Region 'Privates

	End Class

End Namespace