﻿Namespace Global.System.Windows.Forms

	''' <summary>Organisationseinheit mehrerer ToolStripMenuItem, stellt sicher, dass maximal eines checked ist</summary>
	Public Class RadioMenuBehavior

		Public Event CheckedChanged As EventHandler
		Public ReadOnly Items As ToolStripMenuItem()
		Private _Checked As ToolStripMenuItem

		Public Sub New(ParamArray items As ToolStripMenuItem())
			Me.Items = items
			Me.Items.ForEach(Sub(rb) rb.CheckOnClick = False)
			Me.Items.ForEach(Sub(rb)
									  rb.Checked = False
									  AddHandler rb.Click, Sub(s, e) Checked = DirectCast(s, ToolStripMenuItem)
								  End Sub)
		End Sub

		Public Property Checked() As ToolStripMenuItem
			Get
				Return _Checked
			End Get
			Set(value As ToolStripMenuItem)
				If TrySetChecked(value) Then RaiseEvent CheckedChanged(Me, EventArgs.Empty)
			End Set
		End Property

		Public Function TrySetChecked(value As ToolStripMenuItem) As Boolean
			If value Is _Checked Then Return False
			If value IsNot Nothing AndAlso Not Items.Contains(value) Then Throw New ArgumentException("value muss ein bekannter ToolstripMenuItem sein, oder Nothing")
			If _Checked IsNot Nothing Then _Checked.Checked = False
			_Checked = value
			If _Checked IsNot Nothing Then _Checked.Checked = True
			Return True
		End Function

		''' <summary>wählt das erste MenuItem, was auf predicate matcht (oder keines) </summary>
		Public Sub Choose(predicate As Func(Of ToolStripMenuItem, Boolean))
			TrySetChecked(Items.FirstOrDefault(predicate))
		End Sub

	End Class


End Namespace