﻿Imports System.Runtime.CompilerServices

Friend Class App

    Public Shared Property PersistanceModeXml As Boolean = False
#If DEBUG Then
    Public Shared Property DtsXmlFile As String = "..\..\Daten\dts.xml"
#Else
    Public Shared Property DtsXmlFile As String = $"{Application.StartupPath}\dts.xml"
#End If

End Class

Friend Module LocalHelper

    Private _ToDelete As New Dictionary(Of DataRow, ValueTuple(Of DataTable, Integer, Object()))
    Private _DeleteReason As String = ""

    ''' <summary>Erlaubt das Löschen von Zeilen aus einem DGV und somit auch aus der Datenbank nach vorherigem Speichern und Rückfrage</summary>
    ''' <param name="dgv">Das DataGridView</param>
    ''' <param name="e">Die KeyEventArgs</param>
    <Extension>
    Friend Sub DeleteRows(dgv As DataGridView, e As KeyEventArgs, Optional LogDeletion As Boolean = True)
        Dts.SaveDts()
        dgv.AllowUserToDeleteRows = True
        _ToDelete = New Dictionary(Of DataRow, ValueTuple(Of DataTable, Integer, Object()))
        If LogDeletion Then RequestDeleteReason()
        For Each rw In dgv.SelectedRowsX
            Dim boundrow = TryCast(rw.DataBoundItem, DataRowView).Row
            If boundrow IsNot Nothing Then
                Dim recordID = If(boundrow.Table.Columns(0).ColumnName = "ID", CInt(boundrow.Item(0).ToString), 0)
                _ToDelete.Add(boundrow, (boundrow.Table, recordID, boundrow.ItemArray))
            End If
        Next
        If LogDeletion Then AddHandler dgv.UserDeletedRow, AddressOf dgv_UserDeletedRow
        e.Handled = msgQuestionWarning("Die markierten Zeilen werden endgültig gelöscht, fortfahren?") = DialogResult.No
        Dts.SaveDts()
    End Sub

    Private Sub dgv_UserDeletedRow(sender As Object, e As DataGridViewRowEventArgs)
        For Each kvp In _ToDelete
            If kvp.Key.RowState = DataRowState.Deleted Then
                Dim logger = New DataLogger(If(_DeleteReason = "", "Datensatz gelöscht - kein Grund angegeben", $"Datensatz gelöscht: {_DeleteReason}"), kvp.Value)
                logger.SaveDeletedRow()
            End If
        Next
        Dts.SaveDts()
    End Sub

    Private Sub RequestDeleteReason()
        Using dlg As New dlgTextInput
            dlg.Text = "Löschgrund"
            If dlg.ShowDialog = DialogResult.OK Then _DeleteReason = dlg.tbInput.Text
        End Using
    End Sub

End Module

Friend Module Daten

#Region "CreateNew"

    ''' <summary>Öffnet einen Dialog zur Neuanlage eines Datensatzes</summary>
    <Extension>
    Public Function ShowCreateNewDialog(dgv As DataGridView, editDlg As Form) As Boolean
        Dim bs = DirectCast(dgv.DataSource, BindingSource)
        Try
            If bs.EditNew(editDlg) = DialogResult.OK Then Return True
        Catch ex As Exception
            msgEx(ex)
            bs.CancelEdit()
            Return False
        End Try
        Return False
    End Function

    ''' <summary>Öffnet einen Dialog zur Neuanlage eines Datensatzes</summary>
    <Extension>
    Public Function ShowCreateNewDialog(Of T As {Form, New})(BS As BindingSource) As Boolean
        Dim logger = New DataLogger("Erstanlage")
        Try
            If BS.EditNew(Of T) = DialogResult.OK Then
                Dim newRw = BS.DataTable.RowsX.FirstOrDefault(Function(x) x.RowState = DataRowState.Added)
                logger.LogNewDataRow(newRw)
                Return True
            End If
        Catch Ex As Exception
            msgEx(Ex)
            BS.CancelEdit()
            Return False
        End Try
        Return False
    End Function

    ''' <summary>Öffnet einen Dialog zur Neuanlage eines Datensatzes</summary>
    <Extension>
    Public Function ShowCreateNewDialogTpl(Of T As {Form, New})(BS As BindingSource) As (Boolean, DataRow)
        Dim logger = New DataLogger("Erstanlage")
        Try
            If BS.EditNew(Of T) = DialogResult.OK Then
                Dim newRw = BS.DataTable.RowsX.FirstOrDefault(Function(x) x.RowState = DataRowState.Added)
                logger.LogNewDataRow(newRw)
                Return (True, newRw)
            End If
        Catch Ex As Exception
            msgEx(Ex)
            BS.CancelEdit()
            Return (False, Nothing)
        End Try
        Return (False, Nothing)
    End Function

#End Region 'CreateNew

#Region "Edit"

    ''' <summary>Öffnet einen Dialog zum Bearbeiten des gewählten Datensatzes</summary>
    <Extension>
    Friend Sub ShowEditDialog(dgv As DataGridView, editdlg As Form)
        Dim bs = DirectCast(dgv.DataSource, BindingSource)
        Dim dr = DirectCast(bs.Current, DataRowView).Row
        Dim logger As New DataLogger("Bearbeitet", dr)
        If bs.EditCurrent(editdlg) = DialogResult.OK Then logger.CompareValuesAndStoreChanges()
    End Sub

    ''' <summary>Öffnet einen Dialog zum Bearbeiten des gewählten Datensatzes</summary>
    <Extension>
    Friend Function ShowEditDialog(Of T As {Form, New})(BS As BindingSource, Optional UseDataLogger As Boolean = True) As DialogResult
        If BS.Count = 0 Then
            msgError("Bitte einen Eintrag auswählen")
            Return DialogResult.Cancel
        End If
        Dim dr = DirectCast(BS.Current, DataRowView).Row
        Dim logger = If(UseDataLogger, New DataLogger("Bearbeitet", dr), Nothing)
        If BS.EditCurrent(Of T) = DialogResult.OK Then
            If UseDataLogger Then logger.CompareValuesAndStoreChanges()
        End If
        Return DialogResult.OK
    End Function

#End Region 'Edit

    ''' <summary>Öffnet einen Dialog zur Änderung mehrerer Datensätze gleichzeitig</summary>
    <Extension>
    Friend Sub ShowMultiEditDialog(Of T As {Form, New})(dgv As DataGridView)
        Dim bs = DirectCast(dgv.DataSource, BindingSource)
        Dim drs = dgv.SelectedRowsX.Select(Function(rw) DirectCast(rw.DataBoundItem, DataRowView).Row).ToArray
        If dgv.SelectedRows.Count > 0 Then
            Dim selRows = dgv.storeDgvSelectedRows
            Dim logger As New DataLogger("Schnelländerung", drs)
            Using dlg = New T
                If dlg.EditMulti(dgv) = DialogResult.OK Then logger.CompareValuesAndStoreChanges()
            End Using
            dgv.MultiSelect = True
            dgv.restoreDgvSelectedRows(selRows)
        Else
            msgInformation("Bitte mind. 1 Zeile auswählen!")
        End If
    End Sub

End Module