﻿Public Class DataLogger

    Private Shared _LogType As String
    Private Shared _OldValuesBeforeEdit As Dictionary(Of DataRow, Object())
    Private Shared _ValuesToSaveBeforeDelete As Object()
    Private Shared _TableToDeleteFrom As DataTable
    Private Shared _IdFromRowToDelete As Integer
    Private Shared _NewLogRow As DataLogRow

    ''' <summary>Spezielle Sub New() für das Loggen beim Löschen von DataRows. Bis der UserDeletedRow-Handler eines
    ''' DataGridViews anspringt, hat die DataRow bereits RowsState.Deleted und beinhaltet keine Values mehr, deshalb
    ''' muss an dieser Stelle mit ValueTuple gearbeitet werden.
    ''' </summary>
    ''' <param name="Typ">Bearbeitungstyp ("Datensatz gelöscht")</param>
    ''' <param name="tpl">ValueTuple mit Tabelle, Datensatz-ID und Datensatz-ItemArray</param>
    Public Sub New(Typ As String, tpl As ValueTuple(Of DataTable, Integer, Object()))
        _LogType = Typ
        _TableToDeleteFrom = tpl.Item1
        _IdFromRowToDelete = tpl.Item2
        _ValuesToSaveBeforeDelete = tpl.Item3
    End Sub

    ''' <summary>New-Methode für das Bearbeiten von Datensätzen</summary>
    Public Sub New(Typ As String, ParamArray Drs As DataRow())
        _LogType = Typ
        _OldValuesBeforeEdit = New Dictionary(Of DataRow, Object())
        Drs.ForEach(Sub(dr) _OldValuesBeforeEdit.Add(dr, dr.ItemArray))
    End Sub

    ''' <summary>New-Methode ausschließlich zur Erstanlage von Datensätzen</summary>
    Public Sub New(TextErstanlage As String)
        _LogType = TextErstanlage
    End Sub

    ''' <summary>Erzeugt einen Eintrag in der DataLog-Tabelle, wann und durch welchen Benutzer der Datensatz erstellt wurde</summary>
    Friend Sub LogNewDataRow(dr As DataRow)
        Dim DicNegativeRecordID As New Dictionary(Of DataLogRow, DataRow)
        Dim recordID = If(dr.Table.Columns(0).ColumnName = "ID", CInt(dr.Item(0).ToString), 0)
        Dim rwNew = Dts.DataLog.AddDataLogRow(dr.Table.TableName, Environment.UserName, Date.Now, recordID, "", "", BuildItemArrayString(dr.Table.Columns, dr.ItemArray), _LogType)
        If recordID < 0 Then
            Dts.SaveDts() 'Speichern, damit die ID aus der Datenbank übernommen wird
            rwNew.RecordID = CInt(dr.Item(0).ToString)
            Dts.SaveDts() 'Logger-Eintrag mit geänderter ID speichern
        End If
    End Sub

    ''' <summary>Änderungsverfolgung: Vergleicht die Werte aller DataRows in _OldValues mit ggf. geänderten Werten. Gibt es Änderungen, werden diese in der Tabelle
    ''' DataLog mit zusätzlichen Informationen abgespeichert</summary>
    Friend Sub CompareValuesAndStoreChanges()
        Dim DicNegativeRecordID As New Dictionary(Of DataLogRow, DataRow)
        For Each kvp In _OldValuesBeforeEdit
            Dim dr = kvp.Key
            Dim oldValues = kvp.Value
            Dim columns = dr.Table.Columns
            Dim recordID = If(dr.Table.Columns(0).ColumnName = "ID", CInt(dr.Item(0).ToString), 0)
            For iCol = 0 To columns.Count - 1
                Dbg($"DataLogger: {dr.Table.TableName}:{dr.Table.Columns(iCol).ColumnName} Alter Wert: {oldValues(iCol)}")
                Dbg($"DataLogger: {dr.Table.TableName}:{dr.Table.Columns(iCol).ColumnName} Neuer Wert: {dr.Item(iCol)}")
                If dr.Item(iCol).ToString <> oldValues(iCol).ToString Then
                    _NewLogRow = Dts.DataLog.AddDataLogRow(dr.Table.TableName, Environment.UserName, Date.Now, recordID, columns(iCol).ColumnName, oldValues(iCol).ToString, dr.Item(iCol).ToString, _LogType)
                    If recordID < 0 Then DicNegativeRecordID.Add(_NewLogRow, dr)
                    Dts.SaveDts()
                End If
            Next
        Next
        If DicNegativeRecordID.Count > 0 Then
            Dts.SaveDts() 'Speichern ist hier notwendig, damit negative ID's in positive aus der DB umgewandelt werden.
            For Each kvp In DicNegativeRecordID
                kvp.Key.RecordID = CInt(kvp.Value.Item(0).ToString)
            Next
        End If
    End Sub

    ''' <summary>Loggt das Löschen einer DataRow</summary>
    Friend Sub SaveDeletedRow()
        DeleteOldLogs(_TableToDeleteFrom, _IdFromRowToDelete)
        Dts.DataLog.AddDataLogRow(_TableToDeleteFrom.TableName, Environment.UserName, Date.Now, _IdFromRowToDelete, "", BuildItemArrayString(_TableToDeleteFrom.Columns, _ValuesToSaveBeforeDelete), "", _LogType)
        Dts.SaveDts()
    End Sub

    ''' <summary>Löscht evtl. vorhandene Logging-Einträge der zu löschenden DataRow, damit die DataLog-Tabelle nicht weiter wächst</summary>
    Private Sub DeleteOldLogs(tbl As DataTable, recordID As Integer)
        Dim ToDelete As New List(Of DataRow)
        Dts.DataLog.Where(Function(x) x.Tablename = tbl.TableName AndAlso x.RecordID = recordID).ForEach(Sub(rw) ToDelete.Add(rw))
        ToDelete.ForEach(Sub(rw) rw.Delete())
    End Sub

    ''' <summary>Erstellt einen String mit Spaltenname:Value anhand ColumnCollection und Values()</summary>
    Private Function BuildItemArrayString(cols As DataColumnCollection, values As Object()) As String
        Dim strBuilder As New System.Text.StringBuilder
        For iCol = 0 To cols.Count - 1
            strBuilder.Append($"{cols(iCol).ColumnName}:{values(iCol).ToString} | ")
        Next
        Return strBuilder.ToString
    End Function

End Class
