﻿Imports System.Text, System.Text.RegularExpressions, System.Runtime.CompilerServices, System.Diagnostics

Public Enum Provider
    Access = 1
    MsSQL = 2
    MySQL = 3
End Enum

Public Module DbGeneratorMod

    Private _provider As Provider
    Private _sCon As String
    Private _displayOnly As Boolean

    Public Sub ShowSql(dts As DataSet)
        InitQuotes("` `")
        _provider = Provider.MySQL
        Using dlg As New dlgRtxt
            Dim rtb = dlg.RichTextBox1
            rtb.Lines = GetBuildString(dts).ToArray
            dlg.ShowDialog()
        End Using
    End Sub

    Private Function GetBuildString(dts As DataSet) As List(Of String)
        Dim lines As New List(Of String)
        Dim lstStr As New List(Of String)
        Dim addRuleAction As Action(Of String, Rule) = Sub(rulename As String, rule As Rule)
                                                           If _provider = Provider.MsSQL Then 'MsSQL verlangt an der Stelle ausdrücklich ein "NO ACTION", bei anderen Providern kann das wegbleiben
                                                               If rule = Rule.None Then lstStr.Add($"ON {rulename} NO ACTION")
                                                               Return
                                                           Else
                                                               If rule = Rule.None Then Return
                                                           End If
                                                           lstStr.Add($"ON {rulename} {If(rule = Rule.Cascade, "CASCADE", If(rule = Rule.SetDefault, "SET DEFAULT", "SET NULL"))}")
                                                       End Sub
        Dim uniqueName As New Uniquifier
        Dim colHasNoExpression As Func(Of DataColumn, Boolean) = Function(col) col.Expression.HasNoValue
        Dim tbls = dts.GetRankedTables.Where(Function(tbl) tbl.PrimaryKey.Length > 0).ToArray 'Tabellen ohne PrimKey-Column überspringen
        For Each tbl As DataTable In tbls
            If tbl.Columns.Count = 0 Then Continue For
            lstStr.Clear()
            Select Case _provider
                Case Provider.Access : lstStr.Add($"CREATE TABLE {tbl.TableName.Quote} (") 'IF NOT EXISTS testen
                Case Provider.MySQL : lstStr.Add($"CREATE TABLE {tbl.TableName.Quote} (") 'IF NOT EXISTS testen
                Case Provider.MsSQL : lstStr.Add($"CREATE TABLE {tbl.TableName.Quote} (")
            End Select
            For Each col In tbl.Columns.Cast(Of DataColumn).Where(colHasNoExpression)
                Dim colStr = GetColString(col)
                If colStr IsNot Nothing Then lstStr.Add(colStr)
            Next
            For Each uConstraint In tbl.Constraints.OfType(Of UniqueConstraint)
                With uConstraint
                    If .IsPrimaryKey Then
                        lstStr.Add($"CONSTRAINT {$"PK_{ .Table.TableName}".Quote} PRIMARY KEY")
                    Else
                        lstStr.Add($"CONSTRAINT {uniqueName("U_", .Table.TableName).Quote} UNIQUE")
                    End If
                    lstStr.Add(.Columns.ColumngroupString)
                End With
                lstStr.Add(",")
            Next
            For Each fkConstraint In tbl.Constraints.OfType(Of ForeignKeyConstraint)
                With fkConstraint
                    lstStr.Add($"CONSTRAINT {uniqueName("FK_", .RelatedTable.TableName, "_", .Table.TableName).Quote}")
                    lstStr.Add($"FOREIGN KEY { .Columns.ColumngroupString}")
                    lstStr.Add($"REFERENCES { .RelatedTable.TableName.Quote}{ .RelatedColumns.ColumngroupString}")
                    addRuleAction("UPDATE", .UpdateRule)
                    addRuleAction("DELETE", .DeleteRule)
                End With
                lstStr.Add(",")
            Next
            If _provider = Provider.MySQL Then
                lstStr.Poke(");")
            Else
                lstStr.Poke(")")
            End If
            lines.Add(" ".Between(lstStr))
            lines.Add("")
        Next
        Dim allUni = From tbl In tbls
                     From uConstraint In tbl.Constraints.OfType(Of UniqueConstraint)
                     Where uConstraint.Columns.All(colHasNoExpression)
                     Select tbl = tbl.TableName, columns = $"{uConstraint.Columns.ColumngroupString}"
        Dim allFK = From tbl In tbls
                    From fkConstraint In tbl.Constraints.OfType(Of ForeignKeyConstraint)
                    Where fkConstraint.RelatedColumns.All(colHasNoExpression)
                    Select tbl = fkConstraint.RelatedTable.TableName, columns = $"{fkConstraint.RelatedColumns.ColumngroupString}"
        For Each idx In allUni.Concat(allFK).Distinct
            If _provider = Provider.MySQL Then
                lines.Add($"CREATE INDEX {uniqueName($"index{idx.tbl}").Quote} ON {idx.tbl.Quote}{idx.columns};")
            Else
                lines.Add($"CREATE INDEX {uniqueName($"index{idx.tbl}").Quote} ON {idx.tbl.Quote}{idx.columns}")
            End If
        Next
        Return lines
    End Function

    Private Function ColOptionString(col As DataColumn, dbDependentName As String) As String
        Dim colStrBld As New StringBuilder
        Dim strAutoincrement = ""
        Select Case _provider
            Case Provider.Access : strAutoincrement = "COUNTER" '(1, 1)-was hat es damit auf sich??
            Case Provider.MySQL : strAutoincrement = $"{dbDependentName} AUTO_INCREMENT"
            Case Provider.MsSQL : strAutoincrement = "INT IDENTITY"
        End Select
        Dim strNotNull = " NOT NULL"
        Dim strNull = " NULL"
        Dim strUnique = " UNIQUE"
        If col.AutoIncrement Then colStrBld.Append(strAutoincrement) : dbDependentName = ""
        If Not col.AutoIncrement AndAlso col.Unique Then colStrBld.Append(strUnique) 'Unique nur hinzufügen, wenn Col nicht sowieso ID(mit Autoinkrement) ist
        If Not col.AllowDBNull Then colStrBld.Append(strNotNull)
        Return $"{col.ColumnName.Quote} {dbDependentName}{colStrBld},"
    End Function

    Public Function GetColString(col As DataColumn) As String
        Dim dbDependentName As String = ""
        Select Case col.DataType
            Case GetType(Boolean)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "TINYINT"
                    Case Provider.MsSQL : dbDependentName = "bit"
                    Case Provider.Access : dbDependentName = "Bit"
                End Select
            Case GetType(Byte)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "TINYINT UNSIGNED"
                    Case Provider.MsSQL : dbDependentName = Nothing
                    Case Provider.Access : dbDependentName = "Byte"
                End Select
            Case GetType(Byte())
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "varbinary"
                    Case Provider.MsSQL : dbDependentName = "image"
                    Case Provider.Access : dbDependentName = "LongBinary"
                End Select
            Case GetType(Char)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "NCHAR"
                    Case Provider.MsSQL : dbDependentName = "nchar"
                    Case Provider.Access : dbDependentName = "VarChar"
                End Select
            Case GetType(Date)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "DATETIME"
                    Case Provider.MsSQL : dbDependentName = "datetime"
                    Case Provider.Access : dbDependentName = "DateTime"
                End Select
            Case GetType(DateTimeOffset)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = Nothing
                    Case Provider.MsSQL : dbDependentName = Nothing
                    Case Provider.Access : dbDependentName = Nothing
                End Select
            Case GetType(Decimal)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "DECIMAL(10,2)"
                    Case Provider.MsSQL : dbDependentName = "decimal(18,2)"
                    Case Provider.Access : dbDependentName = "Double"
                End Select
            Case GetType(Double)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "DOUBLE"
                    Case Provider.MsSQL : dbDependentName = "float"
                    Case Provider.Access : dbDependentName = "Double"
                End Select
            Case GetType(Guid)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = Nothing
                    Case Provider.MsSQL : dbDependentName = "uniqueidentifier"
                    Case Provider.Access : dbDependentName = "GUID"
                End Select
            Case GetType(Short)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "SMALLINT"
                    Case Provider.MsSQL : dbDependentName = "smallint"
                    Case Provider.Access : dbDependentName = "Short"
                End Select
            Case GetType(Integer)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "INT"
                    Case Provider.MsSQL : dbDependentName = "int"
                    Case Provider.Access : dbDependentName = "Long"
                End Select
            Case GetType(Long)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "BIGINT"
                    Case Provider.MsSQL : dbDependentName = "bigint"
                    Case Provider.Access : dbDependentName = Nothing
                End Select
            Case GetType(Object)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = Nothing
                    Case Provider.MsSQL : dbDependentName = Nothing
                    Case Provider.Access : dbDependentName = Nothing
                End Select
            Case GetType(SByte)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "TINYINT"
                    Case Provider.MsSQL : dbDependentName = "tinyint"
                    Case Provider.Access : dbDependentName = Nothing
                End Select
            Case GetType(Single)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "FLOAT"
                    Case Provider.MsSQL : dbDependentName = "real"
                    Case Provider.Access : dbDependentName = "Single"
                End Select
            Case GetType(String)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = If(col.MaxLength = -1, "VARCHAR(255)", "TEXT") 'Mediumtext: 16,777,215 Chars, Longtext: 4,294,967,295 Chars
                    Case Provider.MsSQL : dbDependentName = If(col.MaxLength = -1, "text", If(col.MaxLength <= 8000, $"varchar({col.MaxLength})", "text"))
                    Case Provider.Access : dbDependentName = If(col.MaxLength = -1, "LongText", If(col.MaxLength > 64000, "LongText", "VarChar"))
                End Select
            Case GetType(TimeSpan)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "TIME"
                    Case Provider.MsSQL : dbDependentName = Nothing
                    Case Provider.Access : dbDependentName = Nothing
                End Select
            Case GetType(UShort)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "SMALLINT UNSIGNED"
                    Case Provider.MsSQL : dbDependentName = Nothing
                    Case Provider.Access : dbDependentName = Nothing
                End Select
            Case GetType(UInteger)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "MEDIUMINT UNSIGNED"
                    Case Provider.MsSQL : dbDependentName = Nothing
                    Case Provider.Access : dbDependentName = Nothing
                End Select
            Case GetType(ULong)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "BIGINT UNSIGNED"
                    Case Provider.MsSQL : dbDependentName = Nothing
                    Case Provider.Access : dbDependentName = Nothing
                End Select
        End Select
        If dbDependentName Is Nothing Then
            msgInformation($"Der Datentyp {col.DataType} für Spalte {col.ColumnName} ist nicht implementiert, wird übersprungen!")
            Return Nothing
        End If
        Return ColOptionString(col, dbDependentName)
    End Function

End Module

Module CreateTableHelpers

#Region "IList"

    <Extension(), DebuggerStepThrough()>
    Public Sub Add(Of T)(ByVal subj As ICollection(Of T), ParamArray itms() As T)
        For Each itm In itms : subj.Add(itm) : Next
    End Sub

    ''' <summary>Bsp: setze vorletztes Element: list.Poke(-1, "huhu")</summary>
    <Extension(), DebuggerStepThrough()>
    Public Sub Poke(Of T)(ByVal subj As IList(Of T), ByVal value As T, Optional ByVal index As Integer = 0)
        subj(index + subj.Count - 1) = value
    End Sub

#End Region 'IList

    Dim _InnerQuotes As String
    Dim _OuterQuotes As String()
    Dim _Parenthesis As String() = {"( ", " )"}

    Sub InitQuotes(quotes As String)
        _OuterQuotes = quotes.Split
        _InnerQuotes = $"{_OuterQuotes(1)},{_OuterQuotes(0)}"
    End Sub

    <Extension()>
    Function ColumngroupString(cols As DataColumn()) As String
        Return _InnerQuotes.Between(cols).Between(_OuterQuotes).InParenthesis
    End Function

    <Extension()>
    Public Function InParenthesis(ByVal Subj As String) As String
        Return Subj.Between(_Parenthesis)
    End Function

    <Extension()>
    Public Function Quote(ByVal Subj As String) As String
        Return Subj.Between(_OuterQuotes)
    End Function

End Module

Public Class Uniquifier : Inherits Dictionary(Of String, Integer)

    Private Shared _rgxLastDigits As New Regex(" \ d + $", RegexOptions.RightToLeft)

    ''' <summary> appends a number at baseword, to ensure, that its unique in the previousUniques-List  </summary> 
    Public Function Uniquify(baseWord As String, previousUniques As IEnumerable(Of String)) As String
        Dim mt = _rgxLastDigits.Match(baseWord)
        If mt.Success Then baseWord = baseWord.Substring(0, mt.Index)
        Dim i = If(mt.Success, Integer.Parse(mt.Value), 0)
        For Each wrd In previousUniques
            mt = _rgxLastDigits.Match(wrd)
            If mt.Success Then wrd = wrd.Substring(0, mt.Index)
            If wrd = baseWord Then i = Math.Max(i, Integer.Parse(mt.Value))
        Next
        Return baseWord & i + 1
    End Function

    ''' <summary> appends a number at baseword, to ensure, that its unique in this instance of Uniquifier </summary> 
    Default Public Overloads ReadOnly Property Item(baseword As String) As String
        Get
            baseword = _rgxLastDigits.Replace(baseword, "")
            Dim i = 0
            If TryGetValue(baseword, i) Then i += 1
            MyBase.Item(baseword) = i
            Return baseword & i
        End Get
    End Property

    ''' <summary> appends a number at baseword, to ensure, that its unique in this instance of Uniquifier </summary> 
    Default Public Overloads ReadOnly Property Item(baseword As String, ParamArray basewordSegments() As String) As String
        Get
            Return Item(baseword & String.Concat(basewordSegments))
        End Get
    End Property

End Class

