﻿Imports System.Runtime.CompilerServices, System.Reflection, System.Collections, System.Drawing, Microsoft.VisualBasic.ControlChars
Friend Module Helpers
	''' <summary> Speichert die Indizes der selektierten DGV-Rows in eine List(Of Integer) </summary>
	<Extension>
	Public Function storeDgvSelectedRows(dgv As DataGridView) As List(Of Integer)
		Dim lstRows As New List(Of Integer)
		For Each rw As DataGridViewRow In dgv.SelectedRows
			lstRows.Add(rw.Index)
		Next
		Return lstRows
	End Function

	''' <summary> Ruft die Indizes von vorher selektierten DGV-Rows aus einer Liste ab und stellt diese wieder her </summary>
	<Extension>
	Public Sub restoreDgvSelectedRows(dgv As DataGridView, lst As List(Of Integer))
		For Each itm In lst
			If dgv.Rows.Count >= lst.Count Then dgv.Rows(itm).Selected = True
		Next
	End Sub

	''' <summary> Erweitertes Error-Handling für DataGridView </summary>
	Public Sub DgvErrorHandling(sender As Object, e As DataGridViewDataErrorEventArgs)
		Dim dgv = DirectCast(sender, DataGridView)
		If Not e.ColumnIndex = -1 Then
			Dim comboCol = TryCast(dgv.Columns(e.ColumnIndex), DataGridViewComboBoxColumn)
			If comboCol IsNot Nothing AndAlso e.Context = (DataGridViewDataErrorContexts.Formatting Or DataGridViewDataErrorContexts.Display) _
				AndAlso TypeOf e.Exception Is ArgumentException Then
				Dim bsPK = DirectCast(comboCol.DataSource, BindingSource)
				If bsPK Is Nothing Then
					Throw New Exception($"{dgv.Name}: FEHLER! Sind alle Bindingsources auf Comboboxen gesetzt?")
					Return
				End If
				Dim tbPK = bsPK.DataTable
				Dim defaultRow = tbPK.Rows.Cast(Of DataRow).FirstOrDefault(Function(rw) 0 = (rw.RowState And (DataRowState.Deleted Or DataRowState.Detached)))
				If defaultRow.NotNull Then
					dgv(e.ColumnIndex, e.RowIndex).Value = defaultRow(comboCol.ValueMember) ' gültigen PK-Wert einsetzen
					Return
				End If
			End If
		End If
		If TypeOf e.Exception Is InvalidConstraintException Then
			msgError($"Der Datensatz kann nicht gelöscht werden, da andere Datensätze mit diesem in Beziehung stehen.")
		Else
			Msg("ColumnIndex: ", e.ColumnIndex, ", Context: ", e.Context, Lf, e.Exception.Message)
		End If
	End Sub

	<Extension()>
	Public Function ColumnsX(dgv As DataGridView) As List(Of DataGridViewColumn)
		Return dgv.Columns.Cast(Of DataGridViewColumn).ToList
	End Function

	<Extension>
    Public Function SelectedRowsX(dgv As DataGridView) As List(Of DataGridViewRow)
        Return dgv.SelectedRows.Cast(Of DataGridViewRow).ToList
    End Function

	Public Sub RegisterDgvSettings(ParamArray dgvs() As DataGridView)
		dgvs.ForEach(Sub(dgv) dgv.RegisterDgvSettings)
	End Sub

	''' <summary>registriert alle relevanten Settings für das DGV, wenn DataBinding genutzt wird</summary>
	<Extension>
	Public Sub RegisterDgvSettings(dgv As DataGridView, Optional CustomCopyOrExport As Boolean = False)
		Dim frm = dgv.FindForm
		dgv.AutoSizeColumnsMode = DataGridViewAutoSizeColumnsMode.None
		dgv.DGVDoubleBuffer
	End Sub

	''' <summary>aktiviert das DoubleBuffering für ein DGV</summary>
	<Extension>
	Public Sub DGVDoubleBuffer(DGV As DataGridView)
		Dim dgvType As Type = DGV.GetType
		Dim pi As PropertyInfo = dgvType.GetProperty("DoubleBuffered", BindingFlags.Instance Or BindingFlags.NonPublic)
		pi.SetValue(DGV, True, Nothing)
	End Sub

	<Extension>
	Public Sub ShowSelectedRowsCounterInToolStripLabel(dgv As DataGridView, tssl As ToolStripLabel)
		tssl.Text = $"markierte Einträge: {dgv.SelectedRows.Count}"
	End Sub

	<Extension>
	Public Sub ShowTotalRowsCounterInToolStripLabel(dgv As DataGridView, tssl As ToolStripLabel)
		tssl.Text = $"Einträge: {dgv.Rows.Count}"
	End Sub

End Module

Public Module MessageBoxes

	''' <summary>zeigt eine Messagebox, wenn die gefilterte Bindingsource keine Elemente enthält</summary>
	<Extension>
	Public Sub MsgNoEntry(bs As BindingSource)
		If bs.Count = 0 Then msgInformation("keine Einträge")
	End Sub

	Public Sub msgDeleteDenied()
		msgInformation("Löschen nicht möglich!")
	End Sub

	Public Function msgOkCancel(Nachricht As String, Titel As String) As DialogResult
		Return MessageBox.Show(Nachricht, Titel, MessageBoxButtons.OKCancel, MessageBoxIcon.Information, Nothing) ', MessageBoxOptions.ServiceNotification)
	End Function
	Public Sub msgInformation(Nachricht As String)
		MessageBox.Show(Nachricht, "Information", MessageBoxButtons.OK, MessageBoxIcon.Information, Nothing) ', MessageBoxOptions.ServiceNotification)
	End Sub

	Public Sub Msg(ParamArray items As Object())
		MessageBox.Show(String.Concat(items))
	End Sub

	Public Sub msgEx(ex As Exception, Optional customInfo As String = Nothing)
		If customInfo Is Nothing Then
			MessageBox.Show(ex.Message, "Ausnahmefehler", MessageBoxButtons.OK, MessageBoxIcon.Error, Nothing) ', MessageBoxOptions.ServiceNotification)
		Else
			MessageBox.Show($"{ex.Message}{Environment.NewLine}{customInfo}", "Ausnahmefehler", MessageBoxButtons.OK, MessageBoxIcon.Error, Nothing) ', MessageBoxOptions.ServiceNotification)
		End If
	End Sub

	Public Sub msgExclamation(Nachricht As String)
		MessageBox.Show(Nachricht, "Wichtige Info", MessageBoxButtons.OK, MessageBoxIcon.Exclamation, Nothing) ', MessageBoxOptions.ServiceNotification)
	End Sub

	Public Sub msgError(Nachricht As String)
		MessageBox.Show(Nachricht, "FEHLER", MessageBoxButtons.OK, MessageBoxIcon.Error, Nothing) ', MessageBoxOptions.ServiceNotification)
	End Sub

	Public Function msgQuestionWarning(Nachricht As String) As DialogResult
		Return MessageBox.Show(Nachricht, "Warnung", MessageBoxButtons.YesNo, MessageBoxIcon.Warning, Nothing) ', MessageBoxOptions.ServiceNotification)
	End Function

	Public Function msgQuestionInfo(Nachricht As String) As DialogResult
		Return MessageBox.Show(Nachricht, "Frage", MessageBoxButtons.YesNo, MessageBoxIcon.Question, Nothing) ', MessageBoxOptions.ServiceNotification)
	End Function
End Module

Public Module Tree

	''' <summary>zeigt den TreeView an</summary>
	<Extension>
	Public Sub MdiNaviShow(ctl As Control)
		If Not ctl Is Nothing Then
			ctl.Visible = True
			ctl.Dock = DockStyle.Left
		End If
	End Sub

	''' <summary>blendet den TreeView aus</summary>
	<Extension>
	Public Sub MdiNaviHide(ctl As Control)
		If Not ctl Is Nothing Then
			ctl.Visible = False
			ctl.Dock = DockStyle.None
		End If
	End Sub
End Module

Public Class ControlStyling

#Region "TextBox"

	''' <summary> Überschreibt das eingestellte "Styling" der TextBox </summary>
	Public Shared Sub StyleTb(frm As ContainerControl)
		Dim getChilds As Func(Of Control, IEnumerable) = Function(ctl) ctl.Controls '.Controls
		For Each ctl In getChilds.All(frm)
			Dim tb = TryCast(ctl, TextBox)
			If tb.NotNull Then
				' man kann hier auch andere Controls stylen, nicht nur DGV
			End If
		Next
	End Sub

#End Region 'TextBox

#Region "ComboBox"

	''' <summary> Überschreibt das eingestellte "Styling" aller ComboBoxen auf der Form </summary>
	Public Shared Sub StyleCmbList(frm As ContainerControl)
		Dim getChilds As Func(Of Control, IEnumerable) = Function(ctl) ctl.Controls '.Controls
		For Each ctl In getChilds.All(frm)
			Dim cmb = TryCast(ctl, ComboBox)
			If cmb.NotNull Then
				With cmb
					.DropDownStyle = ComboBoxStyle.DropDownList
					.AutoCompleteSource = AutoCompleteSource.ListItems
					.AutoCompleteMode = AutoCompleteMode.SuggestAppend
				End With
			End If
		Next
	End Sub

	Public Shared Sub StyleCmbList(ParamArray cmbs As ComboBox())
		For Each cmb In cmbs
			With cmb
				.DropDownStyle = ComboBoxStyle.DropDownList
				.AutoCompleteSource = AutoCompleteSource.ListItems
				.AutoCompleteMode = AutoCompleteMode.SuggestAppend
			End With
		Next
	End Sub

	''' <summary> Styled einzelne ComboBoxen mit DropDownStyle = Dropdown</summary>
	Public Shared Sub StyleCmbFreetext(ParamArray cmbs As ComboBox())
		For Each cmb In cmbs
			With cmb
				.DropDownStyle = ComboBoxStyle.DropDown
				.AutoCompleteSource = AutoCompleteSource.ListItems
				.AutoCompleteMode = AutoCompleteMode.SuggestAppend
			End With
		Next
	End Sub

	''' <summary> Styled alle im Form enthaltenen ComboBoxen mit DropDownStyle = Dropdown</summary>
	Public Shared Sub StyleCmbFreetext(frm As ContainerControl)
		Dim getChilds As Func(Of Control, IEnumerable) = Function(ctl) ctl.Controls '.Controls
		For Each ctl In getChilds.All(frm)
			Dim cmb = TryCast(ctl, ComboBox)
			If cmb.NotNull Then
				With cmb
					.DropDownStyle = ComboBoxStyle.DropDown
					.AutoCompleteSource = AutoCompleteSource.ListItems
					.AutoCompleteMode = AutoCompleteMode.SuggestAppend
				End With
			End If
		Next
	End Sub

#End Region 'ComboBox

#Region "DGV"

	''' <summary> Überschreibt das eingestellte "Styling" des DataGridView </summary>
	Public Shared Sub StyleDGV(frm As ContainerControl)
		Dim getChilds As Func(Of Control, IEnumerable) = Function(ctl) ctl.Controls '.Controls
		For Each ctl In getChilds.All(frm)
			Dim grd = TryCast(ctl, DataGridView)
			If grd.NotNull Then
				Dim tbl = DirectCast(grd.DataSource, BindingSource)?.DataTable
				If tbl.NotNull Then
					For Each dgvCol As DataGridViewColumn In grd.Columns
						Dim propName = dgvCol.DataPropertyName
						If propName = "" Then Continue For
						Dim dataCol = tbl.Columns(propName)
						If dataCol Is Nothing Then
							msgError($"Die Spalte {propName} existiert nicht.{Environment.NewLine}DGV:{grd.Name}{Environment.NewLine}Tabelle:{tbl.TableName}{Environment.NewLine}Form:{frm.Name}")
							Return
						End If
						If dataCol.ReadOnly = True Then dgvCol.DefaultCellStyle.BackColor = SystemColors.Menu
						If dataCol.DataType = GetType(Decimal) Then
							With dgvCol.DefaultCellStyle
								.Alignment = DataGridViewContentAlignment.MiddleLeft
								'.Format = "c"
								'.Font = New Font("Courier New", If(.Font, If(grd.DefaultCellStyle.Font, grd.Font)).Size)
							End With
						ElseIf dataCol.DataType = GetType(Double) Then
							With dgvCol.DefaultCellStyle
								.Alignment = DataGridViewContentAlignment.MiddleLeft
								'      .Format = "n2"
								'      .Font = New Font("Courier New", If(.Font, If(grd.DefaultCellStyle.Font, grd.Font)).Size)
							End With
						ElseIf dataCol.DataType = GetType(DateTime) Then
							'With dgvCol.DefaultCellStyle
							'	.Format = "dd.MM.yyyy"
							'End With
						ElseIf dataCol.DataType = GetType(TimeSpan) Then
							With dgvCol.DefaultCellStyle
								.Format = "hh\:mm"
							End With
						ElseIf dataCol.DataType = GetType(Int32) Then
							With dgvCol.DefaultCellStyle
								.Alignment = DataGridViewContentAlignment.MiddleLeft
							End With
						ElseIf dataCol.DataType = GetType(System.String) Then
							With dgvCol.DefaultCellStyle
								.Alignment = DataGridViewContentAlignment.MiddleLeft
							End With
						End If
					Next
				Else
					For Each dgvCol In grd.ColumnsX
						If dgvCol.ValueType = GetType(Decimal) Then dgvCol.DefaultCellStyle.Alignment = DataGridViewContentAlignment.MiddleRight
						If dgvCol.ValueType = GetType(Double) Then dgvCol.DefaultCellStyle.Alignment = DataGridViewContentAlignment.MiddleRight
						If dgvCol.ValueType = GetType(Integer) Then dgvCol.DefaultCellStyle.Alignment = DataGridViewContentAlignment.MiddleRight
					Next
				End If
				With grd
					.AllowUserToResizeRows = False
					.AllowUserToResizeColumns = True
					.RowHeadersWidth = 25
					.Attach.UnFocusedSelectionColor()
					'.CellBorderStyle = DataGridViewCellBorderStyle.SingleHorizontal
					.CellBorderStyle = DataGridViewCellBorderStyle.Raised
					.SelectionMode = DataGridViewSelectionMode.FullRowSelect
					.ShowEditingIcon = False
					.RowHeadersVisible = False
					.BackgroundColor = SystemColors.Window
					'Designer-Aufgaben-Einstellung
					.AllowUserToAddRows = False
					.ReadOnly = True
					.AllowUserToDeleteRows = False
					.AllowUserToOrderColumns = False
				End With
				AddHandler grd.DataError, AddressOf DgvErrorHandling
			End If
		Next
	End Sub

	Public Shared Sub StyleDGVlv(dgv As DataGridView)
		Dim tbl = DirectCast(dgv.DataSource, BindingSource)?.DataTable
		If tbl.NotNull Then
			For Each dgvCol In dgv.ColumnsX
				Dim propName = dgvCol.DataPropertyName
				If propName = "" Then Continue For
				Dim dataCol = tbl.Columns(propName)
				If dataCol.ReadOnly = True Then dgvCol.DefaultCellStyle.BackColor = SystemColors.Menu
				If dataCol.DataType = GetType(Decimal) Then
					With dgvCol.DefaultCellStyle
						.Alignment = DataGridViewContentAlignment.MiddleLeft
						'.Format = "c"
						'.Font = New Font("Courier New", If(.Font, If(grd.DefaultCellStyle.Font, grd.Font)).Size)
					End With
				ElseIf dataCol.DataType = GetType(Double) Then
					With dgvCol.DefaultCellStyle
						.Alignment = DataGridViewContentAlignment.MiddleLeft
						'      .Format = "n2"
						'      .Font = New Font("Courier New", If(.Font, If(grd.DefaultCellStyle.Font, grd.Font)).Size)
					End With
				ElseIf dataCol.DataType = GetType(DateTime) Then
					'With dgvCol.DefaultCellStyle
					'	.Format = "dd.MM.yyyy"
					'End With
				ElseIf dataCol.DataType = GetType(TimeSpan) Then
					With dgvCol.DefaultCellStyle
						.Format = "hh\:mm"
					End With
				ElseIf dataCol.DataType = GetType(Int32) Then
					With dgvCol.DefaultCellStyle
						.Alignment = DataGridViewContentAlignment.MiddleLeft
					End With
				ElseIf dataCol.DataType = GetType(System.String) Then
					With dgvCol.DefaultCellStyle
						.Alignment = DataGridViewContentAlignment.MiddleLeft
					End With
				End If
			Next
		Else
			For Each dgvCol In dgv.ColumnsX
				If dgvCol.ValueType = GetType(Decimal) Then dgvCol.DefaultCellStyle.Alignment = DataGridViewContentAlignment.MiddleRight
				If dgvCol.ValueType = GetType(Double) Then dgvCol.DefaultCellStyle.Alignment = DataGridViewContentAlignment.MiddleRight
				If dgvCol.ValueType = GetType(Integer) Then dgvCol.DefaultCellStyle.Alignment = DataGridViewContentAlignment.MiddleRight
			Next
		End If
		With dgv
			.AllowUserToResizeRows = False
			.AllowUserToResizeColumns = True
			.RowHeadersWidth = 25
			.SelectionMode = DataGridViewSelectionMode.FullRowSelect
			.ShowEditingIcon = False
			.RowHeadersVisible = False
			.BackgroundColor = SystemColors.Window
			'Designer-Aufgaben-Einstellung
			.AllowUserToAddRows = False
			.ReadOnly = True
			.AllowUserToDeleteRows = False
			.AllowUserToOrderColumns = False
		End With
		AddHandler dgv.DataError, AddressOf DgvErrorHandling
	End Sub

	''' <summary> Überschreibt das eingestellte "Styling" des DataGridView. Orientiert sich am ListView-Design </summary>
	Public Shared Sub StyleDGVlv(frm As ContainerControl)
		Dim getChilds As Func(Of Control, IEnumerable) = Function(ctl) ctl.Controls '.Controls
		For Each ctl In getChilds.All(frm)
			Dim grd = TryCast(ctl, DataGridView)
			If grd.NotNull Then
				Dim tbl = DirectCast(grd.DataSource, BindingSource)?.DataTable
				If tbl.NotNull Then
					For Each dgvCol As DataGridViewColumn In grd.Columns
						Dim propName = dgvCol.DataPropertyName
						If propName = "" Then Continue For
						Dim dataCol = tbl.Columns(propName)
						If dataCol.ReadOnly = True Then dgvCol.DefaultCellStyle.BackColor = SystemColors.Menu
						If dataCol.DataType = GetType(Decimal) Then
							With dgvCol.DefaultCellStyle
								.Alignment = DataGridViewContentAlignment.MiddleLeft
								'.Format = "c"
								'.Font = New Font("Courier New", If(.Font, If(grd.DefaultCellStyle.Font, grd.Font)).Size)
							End With
						ElseIf dataCol.DataType = GetType(Double) Then
							With dgvCol.DefaultCellStyle
								.Alignment = DataGridViewContentAlignment.MiddleLeft
								'      .Format = "n2"
								'      .Font = New Font("Courier New", If(.Font, If(grd.DefaultCellStyle.Font, grd.Font)).Size)
							End With
						ElseIf dataCol.DataType = GetType(DateTime) Then
							'With dgvCol.DefaultCellStyle
							'	.Format = "dd.MM.yyyy"
							'End With
						ElseIf dataCol.DataType = GetType(TimeSpan) Then
							With dgvCol.DefaultCellStyle
								.Format = "hh\:mm"
							End With
						ElseIf dataCol.DataType = GetType(Int32) Then
							With dgvCol.DefaultCellStyle
								.Alignment = DataGridViewContentAlignment.MiddleLeft
							End With
						ElseIf dataCol.DataType = GetType(System.String) Then
							With dgvCol.DefaultCellStyle
								.Alignment = DataGridViewContentAlignment.MiddleLeft
							End With
						End If
					Next
				Else
					For Each dgvCol In grd.ColumnsX
						If dgvCol.ValueType = GetType(Decimal) Then dgvCol.DefaultCellStyle.Alignment = DataGridViewContentAlignment.MiddleRight
						If dgvCol.ValueType = GetType(Double) Then dgvCol.DefaultCellStyle.Alignment = DataGridViewContentAlignment.MiddleRight
						If dgvCol.ValueType = GetType(Integer) Then dgvCol.DefaultCellStyle.Alignment = DataGridViewContentAlignment.MiddleRight
					Next
				End If
				With grd
					.AllowUserToResizeRows = False
					.AllowUserToResizeColumns = True
					.RowHeadersWidth = 25
					.SelectionMode = DataGridViewSelectionMode.FullRowSelect
					.ShowEditingIcon = False
					.RowHeadersVisible = False
					.BackgroundColor = SystemColors.Window
					'Designer-Aufgaben-Einstellung
					.AllowUserToAddRows = False
					.ReadOnly = True
					.AllowUserToDeleteRows = False
					.AllowUserToOrderColumns = False
				End With
				AddHandler grd.DataError, AddressOf DgvErrorHandling
			End If
		Next
	End Sub

#End Region 'DGV

End Class